import __builtin__

import docutils.parsers.rst

import PyDS.Tool

# This conversion function reads identity
def identity(argument):
	return argument

# This function cleans option names from their type tag
def cleanOptions(hash):
	res = {}
	for k in hash.keys():
		p = k.find('__')
		if p >= 0: v = k[:p]
		else: v = k
		if type(hash[k]) == type(u''):
			res[v] = hash[k].encode(_PyDS.documentEncoding)
		else:
			res[v] = hash[k]
	return res

# this is a dictionary that learns about new
# keys just by accessing them (they are initialized
# to unchanged parameters, you need to parse them
# yourself). This is used in directives that have dynamic
# options, for example mappings to PyDS objects whose
# options vary (an example would be the pyds-macro
# directive). You can subclass this to get a specialized
# subclass for a directive that uses has_key to decide
# wether an option is valid. Option names can have a
# trailing __<typ> tag that will change what parameter
# type is allowed. Types are from the usual directive
# paramter types.
class LearningDictionary(dict):

	def __getitem__(self, key):
		p = key.find('__')
		convfun = identity
		if p >= 0:
			typ = key[p+2:]
			if hasattr(docutils.parsers.rst.directives, typ):
				convfun = getattr(docutils.parsers.rst.directives, typ)
			else:
				convfun = getattr(__builtin__, typ)
		if self.has_key(key):
			self.__setitem__(key, convfun)
			return dict.__getitem__(self, key)
		else:
			raise KeyError(key)
	
	def __len__(self):
		return 99
	
	def has_key(self, key):
		return 1
	
def pyds_macro( name, arguments, options, content, lineno,
             content_offset, block_text, state, state_machine ):
  """
  This directive calls a PyDS macro with given parameters.
  Since positional parameters are not used with pyds-macro parameters,
  you must pass all positional parameters as options (pass
  by name)! Only macros that return strings, especially HTML
  strings, make sense to be called this way.

  It is used with the the following syntax::
  
  .. pyds-macro:: imageTag
  	:src: /images/xml.gif
	:alt: this is a XML macro
	:border__int: 0

  Usually you use it inline, so you might want to use a
  substitution definition:

  This is how the XML image looks: |xmlImage|

  .. |xmlImage| pyds-macro:: imageTag
  	:src: /images/xml.gif
	:alt: this is the XML image
	:border__int: 0
  
  This directive has one positional parameter, the macro name.
  Options are not defined here, as they are based on macro parameters.
  A result of this is that you sometimes must provide default values
  like the border parameter in the example, because None in this
  case would say that the border attribute is suppressed on the img
  tag.

  Only simple parameters like strings and integers can be passed
  to macro calls, so macros that need more structured input can't
  be called.
  """
  macroname = arguments[0]
  macros = PyDS.Tool.getTool('macros')
  try:
    macro = getattr(macros, macroname)
  except AttributeError:
    error = state_machine.reporter.error( _("Macro %s not found") % macroname,
      docutils.nodes.literal_block(block_text, block_text), line=lineno )
    return [error]
  # if there is a parameter named alt and the method doesn't accept
  # this parameter, remove it from the options - it's a leftover from the
  # directive parser
  if options.has_key('alt'):
    f = macro
    if hasattr(macro, 'im_func'):
      f = macro.im_func
    if hasattr(f, 'func_code'):
      c = f.func_code
      if 'alt' not in c.co_varnames[:c.co_argcount]:
	del options['alt']
  html = macro(**cleanOptions(options))
  raw = docutils.nodes.raw('',html.decode(_PyDS.documentEncoding), format = 'html')
  return [raw]

pyds_macro.arguments = (1,0,0)
pyds_macro.options = LearningDictionary()
pyds_macro.content = 0
  
# Simply importing this module will make the directives available.
docutils.parsers.rst.directives.register_directive( 'pyds-macro', pyds_macro )

def pyds_nugget( name, arguments, options, content, lineno,
             content_offset, block_text, state, state_machine ):
  """
  This directive calls a PyDS nugget with given parameters.
  You must pass all positional parameters as options.

  It is used with the the following syntax::
  
  .. pyds-nugget:: sampleNugget
  	:header: this would be a header
	:body: this might be the body

  This directive has one positional parameter, the nugget name.
  Options are not defined here, as they are based on nugget parameters.
  You can use the same __<type> delimiter as you can with pyds-macro.
  """
  nuggetname = arguments[0]
  nuggets = PyDS.Tool.getTool('nuggets')
  try:
    nugget = getattr(nuggets, nuggetname)
  except AttributeError:
    error = state_machine.reporter.error( _("Nugget %s not found") % nuggetname,
      docutils.nodes.literal_block(block_text, block_text), line=lineno )
    return [error]
  html = nugget(**cleanOptions(options))
  raw = docutils.nodes.raw('',html.decode(_PyDS.documentEncoding), format = 'html')
  return [raw]

pyds_nugget.arguments = (1,0,0)
pyds_nugget.options = LearningDictionary()
pyds_nugget.content = 0
  
# Simply importing this module will make the directives available.
docutils.parsers.rst.directives.register_directive( 'pyds-nugget', pyds_nugget )

def pyds_picture( name, arguments, options, content, lineno,
             content_offset, block_text, state, state_machine ):
  """
  This directive adds a picture from the PictureTool.

  It is used with the the following syntax::
  
  .. pyds-picture:: 1
  	:width: 150
	:height: 150
	:align: left
	:url: http://pyds.muensterland.org/
	:format: rgbjpg
	:interpolation: bicubic
	:bigpage: 640x480

  There is one positional parameter, the picture id. If
  either width or height is left out, but the other is
  given, the given value is used for both. If none is
  given, the actual picture size is used.
  """
  picid = arguments[0]
  pictures = PyDS.Tool.getTool('pictures')
  pic = pictures.getPicture(picid)
  if pic == {}:
    error = state_machine.reporter.error( _("Picture %s not found") % picid,
      docutils.nodes.literal_block(block_text, block_text), line=lineno )
    return [error]
  opts = {
    'align': options.get('align', 'left'),
    'url': options.get('url', None),
    'bigpage': options.get('bigpage', None),
    'picformat': options.get('format', None),
    'interpolation': options.get('interpolation', None),
  }
  size = [options.get('width', None), options.get('height', None)]
  if size == [None, None]:
    size = None
  elif size[0] is not None and size[1] is None:
  	size[1] = size[0]
  elif size[0] is None and size[1] is not None:
  	size[0] = size[1]
  html = pictures.imageTag(pic['id'], tuple(size), **opts)
  raw = docutils.nodes.raw('',html.decode(_PyDS.documentEncoding), format = 'html')
  return [raw]

pyds_picture.arguments = (1,0,0)
pyds_picture.options = {
  'width': docutils.parsers.rst.directives.nonnegative_int,
  'height': docutils.parsers.rst.directives.nonnegative_int,
  'align': identity,
  'url': identity,
  'bigpage': identity,
  'format': identity,
  'interpolation': identity,
}
pyds_picture.content = 0
  
# Simply importing this module will make the directives available.
docutils.parsers.rst.directives.register_directive( 'pyds-picture', pyds_picture )

