"""

Python Desktop Server - Eventlog viewer

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: EventsTool.py,v 1.43 2004/09/27 12:56:58 gb Exp $

import PyDS.Tool
import metakit
import time

class EventsTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("events[time:I,tool:S,msg:S,error:I]")
		db.getas("prefs[verbose:I,frontcount:I]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.events = self.db.view('events')
			self.prefs = self.db.view('prefs')
			if len(self.prefs) == 0:
				self.prefs.append({
					'verbose':0,
					'frontcount':200
				})
			self._commit()
			self.importantTool = 1
			self.shutdownStage = 99
			self.lastPurge = 0
		finally: self._release()
	
	def _backup(self):
		return self._backupTable('prefs')

	def _initopts(self):
		self.showCalendar = 0

	def _log(self, tool, message, error=0):
		if not hasattr(self, 'lock'):
			return # Tool already shut down - no more logging
		try:
			self._acquire()
			now = time.time()
			# only purge every 5 minutes to prevent excessive
			# event log scanning
			if now - self.lastPurge > 300:
				self.lastPurge = now
				oldest = time.time() - 24*3600
				todel = []
				for i in range(0, len(self.events)):
					row = self.events[i]
					if oldest > row.time:
						todel.append(i)
				if len(todel):
					todel.reverse()
					for i in todel:
						self.events.delete(i)
			self.events.append({
				'time':int(time.time()),
				'tool':tool,
				'msg': message,
				'error': error
			})
			self._commit()
			if _PyDS.echoLog:
				print '<%-15s> %s' % (tool, message)
		finally: self._release()
	
	def _logVerbose(self, tool, message):
		if not hasattr(self, 'lock'):
			return # Tool already shut down - no more logging
		if self.getPref('verbose') or _PyDS.verbose:
			self._log(tool, message)

	def _status(self):
		html = PyDS.Tool.StandardTool._status(self)
		html += '<ul>'
		html += _('<li>There are %s events in the eventlog</li>') % len(self.events)
		html += '</ul>'
		return html
	
	# this method returns preferences
	def getPref(self, pref):
		if hasattr(self, 'prefs'):
			return getattr(self.prefs[0], pref, '')
		else: return ''
		
	# this method delivers the number of events
	def getLength(self):
		return len(self.events)
		
	# this method delivers the last 24 hours of events. Each row is built
	# of the time, the tool and the message of an event.
	def getRecentEvents(self, starting=0):
		try:
			self._acquire()
			toshow = []
			# events isn't ordered! We need to iterate over the full set!
			for row in self.events:
				toshow.append([
					row.time,
					row.tool,
					row.msg,
					row.error
				])
			toshow.sort(lambda a, b: -1*cmp(a[0], b[0]))
			min = 0
			max = len(self.events)
			if starting < max:
				min = starting
			if max-min > self.prefs[0].frontcount:
				max = min+self.prefs[0].frontcount
			return toshow[min:max]
		finally: self._release()

	def getPreferencePanels(self):
		return [
			['events_panel', _('Events'),
				_('Here you can configure your event logging')]
		]
	
	def events_panel(self, req):
		html = '<p>%s' % _('Here you can configure your event logging')
		html += '<form action="%s" method="POST">' % req.getUrl('events_redir', tool='events')
		html += '<table border=0 width="100%">'
		html += '<tr><td>%s</td><td>' % _('show how many events')
		html += '<input name="frontcount" size="5" value="%s">' % (
			self.prefs[0].frontcount
		)
		html += '</td></tr>'
		html += '<tr><td>%s</td><td>' % _('verbose logging')
		html += '<select name="verbose">'
		if self.prefs[0].verbose:
			html += '<option value="1" selected>%s' % _('yes')
			html += '<option value="0">%s' % _('no')
		else:
			html += '<option value="1">%s' % _('yes')
			html += '<option value="0" selected>%s' % _('no')
		html += '</select>'
		html += '</td></tr>'
		html += '</table>'
		html += '<input type=submit name="EventsSubmit" value="%s">' % _('save')
		html += '</form>'
		return html
	
	def events_redir(self, req):
		try:
			self._acquire()
			self.prefs[0].verbose = int(req.getFirstValue('verbose'))
			self.prefs[0].frontcount = int(req.getFirstValue('frontcount'))
			if self.prefs[0].frontcount < 50:
				self.prefs[0].frontcount = 50
			self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='events_panel',
				fortool='events'
			)
		finally: self._release()
	
	def index_html(self, req):
		macros = self.getToolByNamespace('macros')
		starting = int(req.getFirstValue('start', 0))
		frontcount = self.getPref('frontcount')
		if starting < 0:
			starting = 0
		html = PyDS.Tool.StringCollector()

		numevents = self.getLength()
		if numevents > frontcount:
			html += '<p align="right">'
			if starting > 0:
				newstart = starting - frontcount
				if newstart < 0:
					newstart = 0
				html += '<a href="%s">%s</a>' % (
					self.getUrl(start=newstart),
					_('%d newer events') % starting
				)
			if numevents > starting + frontcount:
				if starting > 0:
					html += ' | '
				newstart = starting+frontcount
				html += '<a href="%s">%s</a>' % (
					self.getUrl(start=newstart),
					_('%d older events') % (numevents - newstart)
				)
		html += '<table border=0 width="100%">'
		html += '<tr>'
		html += '<th width="10%%" align="left">%s</th>' % _('Time')
		html += '<th width="10%%" align="left">%s</th>' % _('Tool')
		html += '<th width="80%%" align="left">%s</th>' % _('Message')
		html += '</tr>'
		for row in self.getRecentEvents(starting=starting):
			html += '<tr>'
			html += '<td class="whiteboxsmall" valign="top"><nobr>%s</nobr></td>' % macros.Today(now=row[0], withtime=1)
			html += '<td class="whiteboxsmall" valign="top">%s</td>' % row[1]
			if row[3]:
				html += '<td class="whiteboxsmall" valign="top"><span class="errormsg">%s</span></td>' % row[2]
			else:
				html += '<td class="whiteboxsmall" valign="top">%s</td>' % row[2]
			html += '</tr>'
		html += '</table>'
		req.setLocalValue('title', _('system events (%d)') % frontcount)
		return req.renderPage('BaseTemplate', body=str(html))

__desc__ = _("This tool shows you all events from the last 24 hours.")

PyDS.Tool.registerTool(EventsTool, 'events', _('Events'), __desc__, 670)

