"""

Python Desktop Server - Mirror Registry

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: MirrorRegistry.py,v 1.30 2004/02/04 18:08:52 gb Exp $

import metakit

# Registration of installed mirror classes
_Mirrors = []

# register a mirror
def registerMirror(tag, description, klass):
	_Mirrors.append([tag, description, klass, None])

# instantiate mirrors for a tool (used in MirrorTool)
def instantiateMirrors(tool):
	for mirror in _Mirrors:
		mirror[3] = mirror[2](mirror[0], tool, mirror[1])

# get a named mirror
def getMirror(tag):
	for mirror in _Mirrors:
		if mirror[0] == tag:
			return mirror[3]
	return None

# get a list of installed mirrors
def getMirrors():
	liste = []
	for mirror in _Mirrors:
		liste.append(mirror[3])
	return liste

# delete a list of installed mirrors (this is called in the MirrorTool _shutdown)
def delMirrors():
	for mirror in _Mirrors:
		del mirror[3].tool
		del mirror[3].mirrorstatus
		mirror[3] = None

# this class gives the basic functionality and interface of a mirror
class BaseMirror:

	def __init__(self, tag, tool, title):
		self.tag = tag
		self.tool = tool
		self.title = title
		self.toolHash = {}
		self._initdb()
	
	def _initdb(self):
		self.mirrorstatus = self.tool.db.getas(
			"%s[tool:S,uid:S,signature:S]" % self.tag.lower()
		).ordered(2)
		self.tool._commit()
	
	def _backup(self):
		return self.tool._backupTable('%sStatus' % self.tag)

	# usually you overload this one to get better signatures
	# for change detection
	def _signature(self, item):
		return item['guid']
	
	def _resync(self, tool):
		self.toolHash[tool.name] = {}
		tool._resync(self)
		# now iterate over the hash keys and the internal database for
		# this tool to find differences and resync those differences
		todel = []
		toupdate = []
		toadd = []
		idhash = {}
		for row in self.mirrorstatus.select({'tool':tool.name}):
			key = str(row.uid)
			if self.toolHash[tool.name].has_key(key):
				item = tool.getRSSItem(key, fullrss=1)
				idhash[key] = 1
				if self._signature(item) != row.signature:
					idhash[key] = item
					toupdate.append(key)
			else:
				todel.append(key)
		for key in self.toolHash[tool.name].keys():
			if not(idhash.has_key(key)):
				toadd.append(key)
		for key in todel:
			self.deleteItem(tool, key)
		for key in toupdate:
			self.updateItem(tool, key, idhash[key])
		for key in toadd:
			item = tool.getRSSItem(key, fullrss=1)
			self.addItem(tool, key, item)
		return (len(todel), len(toupdate), len(toadd))

	# This method is called after each mirroring call, so that the status
	# of an item is reflected in the status database.
	def _noteItem(self, tool, uniqueId, signature):
		try:
			self.tool._acquire()
			(idx, found) = self.mirrorstatus.locate({'tool':tool.name, 'uid':uniqueId})
			if found:
				self.mirrorstatus[idx].signature = signature
			else:
				self.mirrorstatus.append({
					'tool':tool.name,
					'uid':uniqueId,
					'signature':signature
				})
			self.tool._commit()
		finally: self.tool._release()

	# This method is called after each mirroring call, so that the status
	# of an item is reflected in the status database.
	def _unnoteItem(self, tool, uniqueId):
		try:
			self.tool._acquire()
			(idx, found) = self.mirrorstatus.locate({'tool':tool.name, 'uid':uniqueId})
			if found:
				self.mirrorstatus.delete(idx)
			self.tool._commit()
		finally: self.tool._release()

	# This method is called in the _resync process as a means
	# of telling us what item exist in the tool
	def iHave(self, tool, uniqueId):
		self.toolHash[tool.name][str(uniqueId)] = 1

	# This method checks wether the mirror can work or not
	def canWork(self):
		return 1
	
	# This method returns the number of items in a mirror status
	def getItemCount(self):
		return len(self.mirrorstatus)

	# stub for subclass overloading
	def _addItem(self, tool, uniqueId, item):
		pass

	# stub for subclass overloading
	def _updateItem(self, tool, uniqueId, item):
		pass

	# stub for subclass overloading
	def _deleteItem(self, tool, uniqueId):
		pass

	# This method syncs a single item - either add or update or do nothing
	def syncItem(self, tool, uniqueId, item):
		sig = self._signature(item)
		(idx, found) = self.mirrorstatus.locate({'tool':tool.name, 'uid':uniqueId})
		if found:
			if self.mirrorstatus[idx].signature != sig:
				self._updateItem(tool, uniqueId, item)
				self._noteItem(tool, uniqueId, sig)
		else:
			self._addItem(tool, uniqueId, item)
			self._noteItem(tool, uniqueId, sig)

	# This method adds an item to the mirror.
	def addItem(self, tool, uniqueId, item):
		self._addItem(tool, uniqueId, item)
		self._noteItem(tool, uniqueId, self._signature(item))
	
	# This method updates an item on the mirror.
	def updateItem(self, tool, uniqueId, item):
		self._updateItem(tool, uniqueId, item)
		self._noteItem(tool, uniqueId, self._signature(item))
	
	# This method deletes an item from the mirror.
	def deleteItem(self, tool, uniqueId):
		self._deleteItem(tool, uniqueId)
		self._unnoteItem(tool, uniqueId)
		

