"""

Python Desktop Server - Picture Tool

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: PictureTool.py,v 1.81 2004/10/08 14:17:49 gb Exp $

import os
import re
import time
import glob
import stat
import string
import mimetypes

import Image
import ImageDraw

import PyDS.Tool
import PyDS.EXIF

interpolationAlgorithms = {
	'nearest': (_('Nearest'), Image.NEAREST),
	'bicubic': (_('Bicubic'), Image.BICUBIC),
	'bilinear': (_('Bilinear'), Image.BILINEAR),
	'antialias': (_('Antialias'), Image.ANTIALIAS)
}

pictureFormats = {
	'rgbpng': (_('Truecolor mode, lossless'), 'RGB', 'png'),
	'rgbjpg': (_('Truecolor mode, lossy'), 'RGB', 'jpeg'),
	'palettepng': (_('Palette mode, lossless'), 'P', 'png')
}

pictureTemplate = """#set $thumb = 100
#set $format = $pictures.getFormat($size)
#set $fullwidth = $format.sizex
#set $smallwidth = $format.sizex - 2*$thumb
#set $next = $pictures.getNextPicture($picture.id, $size)
#set $prev = $pictures.getPreviousPicture($picture.id, $size)
<table width=$fullwidth border="0" cellspacing=5 cellpadding=5>
<tr>
#if $prev
<td width=$thumb>$pictures.imageTag($prev.id,($thumb,$thumb),bigpage=$size)</td>
#else
<td width=$thumb>$macros.blind($thumb,$thumb)</td>
#end if
<td width=$smallwidth align="center">
$pictures.imageSizeBar($picture.id, $size)
</td>
#if $next
<td width=$thumb>$pictures.imageTag($next.id,($thumb,$thumb),bigpage=$size)</td>
#else
<td width=$thumb>$macros.blind($thumb,$thumb)</td>
#end if
</tr>
<tr><td width=$fullwidth colspan=3 align="center">
$pictures.imageTag($picture.id, $size, align="center")
</td></tr>
#if $picture.info
<tr><td width=$thumb>$macros.blind($thumb,1)</td>
<td width=$smallwidth class="whiteboxsmall">$picture.info</td>
<td width=$thumb>$macros.blind($thumb,1)</td></tr>
#end if
</table>
"""

def EXIFString(format, value):
	try: return format % _(value)
	except: return ''

def EXIFNumber(format, value):
	try: return format % round(float(value),3)
	except: return ''

def EXIFInteger(format, value):
	try: return format % round(int(value),3)
	except: return ''

def EXIFRatio(format, value):
	try:
		f = string.split(value, '/')
		r = float(f[0]) / float(f[1])
		r = round(r, 3)
		return format % r
	except: return ''

def EXIFSimplifiedRatio(format, value):
	try:
		f = map(int, string.split(value, '/'))
		r = round(float(f[0]) / float(f[1]), 3)
		if f[0] < f[1]:
			f[1] = int(f[1] / f[0])
			f[0] = 1
		else:
			f[0] = round(f[0] / f[1])
			f[1] = 1
		return format % (r, f[0], f[1])
	except: return ''

def EXIFList(format, liste):
	try: return format % tuple(liste)
	except: return ''

EXIFVariables = [
	(_('Camera make'), 'Image Make', EXIFString, '%s'),
	(_('Camera model'), 'Image Model', EXIFString, '%s'),
	(_('Image date'), 'Image DateTime', EXIFString, '%s'),
	(_('Image resolution'), ['EXIF ExifImageWidth', 'EXIF ExifImageLength'], EXIFList, '%s * %s'),
	(_('Flash used'), 'EXIF Flash', EXIFString, '%s'),
	(_('Focal length'), 'EXIF FocalLength', EXIFNumber, '%0.2f mm'),
	(_('Exposure time'), 'EXIF ExposureTime', EXIFSimplifiedRatio, '%0.3f s (%d/%d)'),
	(_('Aperture'), 'EXIF FNumber', EXIFRatio, 'f/%0.1f'),
	(_('ISO equiv.'), 'EXIF ISOSpeedRatings', EXIFInteger, '%d'),
	(_('Exposure Mode'), 'EXIF ExposureProgram', EXIFString, '%s'),
]

class PictureTool(PyDS.Tool.StandardTool):

	def _recachepictures(self):
		try:
			self._acquire()
			self.picturesByPubtime = self.pictures.sort(self.pictures.pubtime)
		finally: self._release()
	
	def _recachecalendar(self, reftime):
		calendar = self.getToolByNamespace('calendar')
		calendar.recache(self, reftime=reftime)
		calendar.recache(self, reftime=reftime, category='')
		try:
			self._acquire()
			for f in self.formats:
				if f.render:
					calendar.recache(self, reftime=reftime, category=f.name)
		finally: self._release()

	def _createdb(klass, db):
		db.getas("pictures[id:I,publish:I,pubtime:F,sizex:I,sizey:I,title:S,desc:S,text:S,format:S,info:S,timestamp:I,interpolation:S]")
		db.getas("prefs[maxid:I,path:S,title:S,desc:S,template:S,interpolation:S]")
		db.getas("formats[id:I,name:S,sizex:I,sizey:I,render:I]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.pictures = self.db.view('pictures').ordered(1)
			self.prefs = self.db.view('prefs')
			if len(self.prefs) == 0:
				self.prefs.append({'maxid':0})
			if self.prefs[0].template == '':
				self.prefs[0].template = pictureTemplate
			if self.prefs[0].interpolation == '':
				self.prefs[0].interpolation = 'bicubic'
			self.formats = self.db.view('formats').ordered(1)
			if len(self.formats) == 0:
				id = 0
				for (x, y) in ((640, 480), (800, 600), (1024, 768), (1280, 1024)):
					self.formats.append({
						'id': id,
						'name': '%dx%d' % (x, y),
						'sizex': x - 20,
						'sizey': y - 20,
						'render': 1
					})
					id += 1
			self._commit()
			self.picturedb = os.path.join(_PyDS.VARDIR, self.name)
			if not(os.path.isdir(self.picturedb)):
				os.makedirs(self.picturedb)
			self._recachepictures()
		finally: self._release()
	
	def _backup(self):
		
		def _splitpubtime(liste):
			for (cn, cv) in liste:
				if cn == 'pubtime':
					(y,m,t) = time.localtime(cv)[:3]
					return [
						('year', y),
						('month', m)
					]

		l = self._backupTable('prefs')
		l.extend(self._backupTable('pictures', sortcols=['pubtime'], groupfn=_splitpubtime))
		l.extend(self._backupTable('formats'))
		return l
	
	def _initopts(self):
		self.rss = 1
	
	def _initactive(self):
		themes = self.getToolByNamespace('themes')
		weblog = self.getToolByNamespace('weblog')
		themes.hookTemplate(
			'PictureTemplate',
			_('picture overview template for the desktop website'),
			'templates/PictureTemplate.tmpl'
		)
		themes.hookTemplate(
			'PictureRendering',
			_('picture overview template for the cloud'),
			'templates/PictureRendering.tmpl'
		)
		navigation = self.getToolByNamespace('navigation')
		navigation._addHook('select', 'pictureindex',
			_('Link to the picture index'),
			self._pictureIndexNavHook)
		weblog._addHook('rssitems', 'blogmarks',
		 	'',
			self._rssItemsForWeblog,
			self._rssItemsForWeblogActive)

	def _rssItemsForWeblog(self, category):
		format = self.getFormatList()[0]['name']
		items = []
		items = self.getRecentPictures(maxnumber=10)
		res = []
		for item in items:
			if item['publish']:
				el = self._translateItemToRSS(item, format)
				res.append(el)
		return res
	
	def _rssItemsForWeblogActive(self, category):
		if category == '': return 1
		else: return 0
	
	def _pictureIndexNavHook(self, qual, value, preset=None):
		if qual == 'html':
			return ''
		elif qual == 'url':
			return self.getStaticUrl()
		else:
			raise ValueError('%s is not a valid qualifier' % qual)
	
	def _initthread(self, server):
		while server.isRunning():
			time.sleep(10)
			if self.queue.length():
				while self.queue.length():
					(jobf, key, data) = self.queue.shift()
					jobf(key, data)
				self.logVerbose(_('work done, going to sleep again'))
	
	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += _('<li>There are %d pictures in the database</li>') % len(self.pictures)
			html += _('<li>There are %d pictures in the render queue</li>') % self.queue.length()
			html += self._threadstatus()
			html += '</ul>'
			return html
		finally: self._release()
	
	def _resync(self, theMirror):
		for pic in self.pictures:
			if pic.publish:
				theMirror.iHave(self, str(pic.id))
	
	def _rerender(self):
		try:
			self._acquire()
			for pic in self.pictures:
				self.renderPicture(pic.id)
		finally: self._release()
	
	def _renderPicture(self, key, data):
		self.renderPictureOnline(key)
	
	def _renderPictureIndex(self, key, data):
		(year, month, day) = key
		self.renderPictureIndexOnline(year, month, day)
	
	# this returns a preference setting
	def getPref(self, pref):
		try:
			self._acquire()
			if hasattr(self.prefs[0], pref):
				return getattr(self.prefs[0], pref)
			else:
				return ''
		finally: self._release()
	
	def getPreferencePanels(self):
		return [
			['pictures_panel', _('Pictures'),
				_('Here you can configure options for your pictures upload helper.')]
		]
	
	def pictures_panel(self, req):
		macros = self.getToolByNamespace('macros')
		html = '<p>%s' % _('Here you can configure your picture upload')
		html += '<form action="%s" method="POST">' % req.getUrl('pictures_redir', tool=self.name)
		html += '<table border=0>'
		html += '<tr><th>%s</th><td>' % _('picture path')
		html += '<input name="path" size=50 value="%s">' % self.getPref('path')
		html += '</td></tr><tr><th>%s</th><td>' % _('Title')
		html += '<input name="title" size=50 value="%s">' % self.getPref('title')
		html += '</td></tr><tr><th>%s</th><td>' % _('Description')
		html += '<textarea name="desc" cols=50 rows=5>%s</textarea>' % macros.quoteHTML(self.getPref('desc'))
		html += '</td></tr><tr><th>%s</th><td>' % _('Template')
		html += '<textarea name="template" cols=50 rows=10>%s</textarea>' % macros.quoteHTML(self.getPref('template'))
		html += '</td></tr><tr><th>%s</th><td>' % _('Sizes')
		html += '<table>'
		html += '<tr><th>%s</th><th>%s</th><th>%s</th></tr>' % (
			_('Name'), _('Render'), _('Real Size')
		)
		for f in self.getFormatList(all=1):
			html += '<tr><td align="right">%s</td>' % f['name']
			if f['render']:
				html += '<td align="center"><input type=checkbox name="render" value="%d" checked></td>' % f['id']
			else:
				html += '<td align="center"><input type=checkbox name="render" value="%d"></td>' % f['id']
			html += '<td class="smallfont">(%d x %d)</td></tr>' % (f['sizex'], f['sizey'])
		html += '</table>'
		html += '</td></tr>'
		html += '<tr><th>%s</th><td>' % _('Interpolation Algorithm')
		html += '<select name="interpolation">'
		for a in interpolationAlgorithms.keys():
			if a == self.getPref('interpolation'):
				html += '<option value="%s" selected>%s' % (a, interpolationAlgorithms[a][0])
			else:
				html += '<option value="%s">%s' % (a, interpolationAlgorithms[a][0])
		html += '</select>'
		html += '</td></tr>'
		html += '<tr><td colspan="2" align="right"><input type="submit" name="PicturesSubmit" value="%s"></td></tr>' % _('Save')
		html += '</table>'
		html += '</form>'
		html += '<form action="%s" method="POST">' % req.getUrl('reset_redir', tool=self.name)
		html += '<p align="right"><input type="submit" name="RestSubmit" value="%s">' % _('reset templates')
		html += '</form>'
		return html
	
	def reset_redir(self, req):
		try:
			self._acquire()
			for pic in self.pictures:
				pic.text = ''
			self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='pictures_panel',
				fortool=self.name
			)
		finally: self._release()
	
	def pictures_redir(self, req):
		try:
			self._acquire()
			self.prefs[0].path = req.getFirstValue('path')
			self.prefs[0].title = req.getFirstValue('title')
			self.prefs[0].desc = req.getFirstValue('desc')
			self.prefs[0].interpolation = req.getFirstValue('interpolation')
			template = req.getFirstValue('template')
			if template:
				self.prefs[0].template = template
			else:
				self.prefs[0].template = pictureTemplate
			render = {}
			for i in req.getAllValues('render'):
				render[int(i)] = 1
			for f in self.formats:
				if render.has_key(f.id):
					f.render = 1
				else:
					f.render = 0
			self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='pictures_panel',
				fortool=self.name
			)
		finally: self._release()
	
	def index_html(self, req):
		html = req.renderPage('PictureTemplate')
		req.setLocalValue('title', self.getPref('title'))
		return req.renderPage('BaseTemplate', body=html)
	
	def edit_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		pic = self.getPicture(id)
		title = pic.get('title','')
		desc = pic.get('desc','')
		text = pic.get('text', '') or self.getPref('template')
		info = pic.get('info','')
		interpolation = pic.get('interpolation', '') or self.getPref('interpolation')
		if text == '':
			text = self.getPref('template')
		html = '<p align="right"><a href="%s">%s</a>' % (
			req.getUrl(),
			_('Click here to go back to the pictures overview')
		)
		html += '<form action="%s" method="POST">' % req.getUrl('edit_redir')
		html += '<input type="hidden" name="id" value="%s">' % id
		html += '<table border=0>'
		html += '<tr><th>%s</th><td><img src="%s"></td></tr>' % (
			_('Image'),
			req.getUrl('thumbnaildb_jpg', id=id)
		)
		html += '<tr><th>%s</th><td><input name="title" value="%s" size=50></td></tr>' % (
			_('Title'),
			macros.quoteHTML(title)
		)
		html += '<tr><th>%s</th><td><select name="publish">' % _('Publish')
		if int(pic.get('publish','0')):
			html += '<option value="0">%s' % _('no')
			html += '<option value="1" selected>%s' % _('yes')
		else:
			html += '<option value="0" selected>%s' % _('no')
			html += '<option value="1">%s' % _('yes')
		html += '</select></td>'
		html += '<tr><th>%s</th><td><select name="format">' % _('Format')
		formats = pictureFormats.keys()
		formats.sort(lambda a, b: cmp(pictureFormats[a][0], pictureFormats[b][0]))
		for f in formats:
			if f == pic['format']:
				html += '<option value="%s" selected>%s' % (
					f, pictureFormats[f][0],
				)
			else:
				html += '<option value="%s">%s' % (
					f, pictureFormats[f][0],
				)
		html += '</select></td>'
		html += '<tr><th>%s</th><td><select name="interpolation">' % _('Interpolation')
		for a in interpolationAlgorithms.keys():
			if a == interpolation:
				html += '<option value="%s" selected>%s' % (
					a, interpolationAlgorithms[a][0]
				)
			else:
				html += '<option value="%s">%s' % (
					a, interpolationAlgorithms[a][0]
				)
		html += '</select></td>'
		html += '<tr><th>%s</th><td><textarea name="desc" cols=50 rows=5>%s</textarea></td></tr>' % (
			_('Description'),
			macros.quoteHTML(desc)
		)
		html += '<tr><th>%s</th><td><textarea name="info" cols=50 rows=5>%s</textarea></td></tr>' % (
			_('Information'),
			macros.quoteHTML(info)
		)
		html += '<tr><th>%s</th><td><textarea name="text" cols=50 rows=10>%s</textarea></td></tr>' % (
			_('Template'),
			macros.quoteHTML(text)
		)
		html += '</table>'
		html += '<p align="right"><input type="submit" name="EditSubmit" value="%s">' % _('Save')
		html += '</form>'
		html += '<p align="right"><a href="%s">%s</a>' % (
			req.getUrl(),
			_('Click here to go back to the pictures overview')
		)
		req.setLocalValue('title', title)
		return req.renderPage('BaseTemplate', body=html)
	
	def edit_redir(self, req):
		try:
			self._acquire()
			id = int(req.getFirstValue('id'))
			(idx, found) = self.pictures.locate({'id':id})
			if found:
				pic = self.pictures[idx]
				pic.title = req.getFirstValue('title')
				pic.desc = req.getFirstValue('desc')
				pic.info = req.getFirstValue('info')
				pic.text = req.getFirstValue('text')
				pic.interpolation = req.getFirstValue('interpolation')
				pic.timestamp = int(time.time())
				if pic.text == self.getPref('template'):
					pic.text = ''
				pic.publish = int(req.getFirstValue('publish'))
				pic.format = req.getFirstValue('format')
				self._commit()
				self._recachecalendar(pic.pubtime)
				if pic.publish:
					self.renderPicture(id)
				else:
					self.unrenderPicture(id)
					self.renderPictureIndexTimeframe(pic.pubtime)
			return req.getUrl('edit_html', id=str(id))
		finally: self._release()
	
	def delete_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		html = '<p><a href="%s">%s</a>' % (
			req.getUrl('delete_redir', id=id),
			_('Click here to delete this picture')
		)
		pic = self.getPicture(id)
		html += '<h3>%s</h3>' % pic['title']
		html += '<p class="smallfont" align="right">'
		html += '%d x %d, %s' % (
			pic['sizex'],
			pic['sizey'],
			macros.Today(now=pic['pubtime'], withtime=1)
		)
		html += '<p><img align="left" src="%s">' % req.getUrl('thumbnaildb_jpg', id=id)
		html += '%s' % pic['desc']
		html += '<p align="right"><a href="%s">%s</a>' % (
			req.getUrl(),
			_('Click here to go back to the pictures overview')
		)
		req.setLocalValue('title', pic['title'])
		return req.renderPage('BaseTemplate', body=html)
	
	def delete_redir(self, req):
		try:
			self._acquire()
			id = int(req.getFirstValue('id'))
			(idx, found) = self.pictures.locate({'id':id})
			if found:
				self.unrenderPicture(id)
				pubtime = self.pictures[idx].pubtime
				self.pictures.delete(idx)
				self._commit()
				self._recachepictures()
				self._recachecalendar(pubtime)
				try: os.unlink(self.pictureName(id))
				except: pass
				try: os.unlink(self.pictureThumbName(id))
				except: pass
				self.renderPictureIndexTimeframe(pubtime)
			return req.getUrl()
		finally: self._release()
	
	def upload_html(self, req):
		macros = self.getToolByNamespace('macros')
		html = '<p>%s' % _('To upload a picture you must fill in at least the title and description for those images you want to add to your database and then press the upload button.')
		html += '<form action="%s" method="POST">' % self.getUrl('upload_redir')
		html += '<table border=0>'
		idx = 0
		imagere = re.compile(r'^image/.*$')
		for fn in os.listdir(self.getPref('path')):
			fullpath = os.path.join(self.getPref('path'), fn)
			mimetype = mimetypes.guess_type(fullpath)[0]
			info = ''
			if mimetype == 'image/jpeg':
				f = open(fullpath, 'rb')
				try: exif = PyDS.EXIF.process_file(f)
				except: exif = {}
				f.close()
				for (title, key, hook, format) in EXIFVariables:
					val = None
					if type(key) == type([]):
						val = []
						for k in key:
							v = exif.get(k)
							if v:
								val.append(v.printable)
					else:
						val = exif.get(key)
						if val:
							val = val.printable
					if val:
						if info:
							info += '<br>\n'
						info += '<b>%s:</b> %s' % (
							title,
							hook(format, val)
						)
			if mimetype and fn[0] != '.' and os.path.isfile(fullpath) and imagere.match(mimetype):
				idx += 1
				html += '<tr><th colspan=2><input type=hidden name="file%d" value="%s">%s</th></tr>' % (idx, fn, fn)
				html += '<tr><th>%s:</th>' % _('Title')
				html += '<td><input name="title%d" size=40></td>' % idx
				html += '</tr>'
				html += '<tr valign="top">'
				html += '<th>%s:<br><img src="%s"></th>' % (_('Description'), self.getUrl('thumbnail_jpg', fn=fn))
				html += '<td><textarea name="desc%d" cols=40 rows=5></textarea></td>' % idx
				html += '</tr>'
				html += '<tr>'
				html += '<th>%s:</th>' % _('Information')
				html += '<td><textarea name="info%d" cols=40 rows=5>%s</textarea></td>' % (idx, macros.quoteHTML(info))
				html += '</tr>'
		if idx:
			html += '<tr><td colspan="2">'
			html += '<input type=submit name="UploadSubmit" value="%s">' % _('upload')
			html += '</td></tr>'
		html += '</table>'
		html += '<input type=hidden name="count" value="%d">' % idx
		html += '</form>'
		html += '<p align="right"><a href="%s">%s</a>' % (
			req.getUrl(),
			_('Click here to go back to the pictures overview')
		)
		req.setLocalValue('title', _('Pictures for Upload'))
		return req.renderPage('BaseTemplate', body=html)

	# This method uploads a file from the given path. You can specify
	# wether you want to delete the original path or wether you
	# want to directly publish the picture. This method returns
	# the id of the uploaded picture.
	def uploadFromPath(self, title, desc, path, info='', delete=0, publish=0):
		try:
			self._acquire()
			im = Image.open(path)
			(sx, sy) = im.size
			self.prefs[0].maxid += 1
			id = self.prefs[0].maxid
			pubtime = time.time()
			format = 'rgbjpg'
			if im.mode == 'P':
				format = 'rgbpng'
			self.pictures.append({
				'id': id,
				'title': title,
				'desc': desc,
				'info': info,
				'publish': publish,
				'pubtime': pubtime,
				'sizex': sx,
				'sizey': sy,
				'format': format,
				'interpolation': self.getPref('interpolation'),
				'timestamp': pubtime
			})
			self._commit()
			self._recachepictures()
			self._recachecalendar(pubtime)
			im = im.convert(pictureFormats[format][1])
			im.save(self.pictureName(id))
			del im
			if delete:
				try: os.unlink(path)
				except: pass
			if publish:
				self.renderPicture(id)
			return id
		finally: self._release()

	def upload_redir(self, req):
		for index in range(int(req.getFirstValue('count'))):
			idx = index + 1
			file = req.getFirstValue('file%d' % idx)
			title = req.getFirstValue('title%d' % idx)
			info = req.getFirstValue('info%d' % idx)
			desc = req.getFirstValue('desc%d' % idx)
			path = os.path.join(self.prefs[0].path, file)
			tfile = '.%s.jpeg' % os.path.splitext(file)[0]
			tpath = os.path.join(self.prefs[0].path, tfile)
			if title and desc and os.path.isfile(path):
				self.uploadFromPath(title, desc, path, info=info, delete=1, publish=0)
				try: os.unlink(tpath)
				except: pass
		return self.getUrl('upload_html')
	
	def thumbnail_jpg(self, req):
		fn = req.getFirstValue('fn')
		path = os.path.join(self.getPref('path'), fn)
		tfn = '.%s.jpeg' % os.path.splitext(fn)[0]
		tpath = os.path.join(self.getPref('path'), tfn)
		if not os.path.exists(tpath) or not os.path.getsize(tpath):
			f = open(tpath, 'wb')
			f.write(self._thumbnail(path, 150, 150))
			f.close()
		else:
			istat = os.stat(path)
			ostat = os.stat(tpath)
			if istat[stat.ST_MTIME] >= ostat[stat.ST_MTIME]:
				f = open(tpath, 'wb')
				f.write(self._thumbnail(path, 150, 150))
				f.close()
		return open(tpath, 'rb').read()
	
	def thumbnaildb_jpg(self, req):
		id = int(req.getFirstValue('id'))
		pic = self.getPicture(id)
		path = self.pictureName(id)
		opath = self.pictureThumbName(id)
		if not(os.path.exists(opath)):
			f = open(opath, 'wb')
			f.write(self._thumbnail(path, 150, 150))
			f.close()
		else:
			istat = os.stat(path)
			ostat = os.stat(opath)
			if istat[stat.ST_MTIME] >= ostat[stat.ST_MTIME]:
				f = open(opath, 'wb')
				f.write(self._thumbnail(path, 150, 150))
				f.close()
		return open(opath, 'rb').read()
	
	def _thumbnail(self, path, sizex, sizey, format='rgbjpg', interpolation='bicubic'):
		try:
			image = Image.open(path)
			if image.mode != 'RGB':
				image = image.convert('RGB')
		except:
			image = Image.new('RGB', (sizex, sizey), (255,255,255))
			imdraw = ImageDraw.Draw(image)
			imdraw.line((0, 0) + image.size, fill=(255,0,0))
			imdraw.line((0, image.size[1], image.size[0], 0), fill=(255,0,0))
			del imdraw
		(ox, oy) = image.size
		(x, y) = self.scaleImage(ox, oy, sizex, sizey)
		if x != ox or y != oy:
			image = image.resize((x, y), interpolationAlgorithms[interpolation][1])
		(mode, ext) = pictureFormats[format][1:]
		ext = string.upper(ext)
		if image.mode != mode:
			image = image.convert(mode)
		of = PyDS.Tool.StringIO()
		image.save(of, ext)
		return of.getvalue()
	
	# this method check wether the spool path is available. It returns
	# the name of the spool path or ''
	def isSpoolPathAvailable(self):
		if self.getPref('path'):
			path = os.path.join(_PyDS.VARDIR, self.getPref('path'))
			if os.path.isdir(path):
				return path
		return ''
	
	# this method returns the x and y an image should have after scaling
	# it to a given bounding box
	def scaleImage(self, x, y, sizex, sizey):
		lf = 1
		if x > sizex and y > sizey:
			lfx = float(sizex) / float(x)
			lfy = float(sizey) / float(y)
			if lfx < lfy:
				lf = lfx
			else:
				lf = lfy
		elif x > sizex:
			lf = float(sizex) / float(x)
		elif y > sizey:
			lf = float(sizey) / float(y)
		return (int(float(x) * lf), int(float(y) * lf))
	
	def _pictureAsHash(self, pic):
		return {
			'id': pic.id,
			'title': pic.title,
			'desc': pic.desc,
			'info': pic.info,
			'publish': pic.publish,
			'pubtime': pic.pubtime,
			'sizex': pic.sizex,
			'sizey': pic.sizey,
			'text': pic.text or self.getPref('template'),
			'format': pic.format or 'rgbjpg',
			'interpolation': pic.interpolation or self.getPref('interpolation'),
			'timestamp': pic.timestamp
		}
	
	# this method returns a list of recent pictures
	def getRecentPictures(self, year=None, month=None, day=None, maxnumber=15, category=None):
		try:
			self._acquire()
			macros = self.getToolByNamespace('macros')
			l = self.picturesByPubtime
			if year and month:
				(min, max) = macros.getMinMaxEpoch(year, month, day)
				l = l.select({'pubtime':min},{'pubtime':max})
			idx = len(l)
			if year and month and day:
				maxnumber = idx
			liste = []
			size = 0
			if category:
				size=int(category)
			while idx and maxnumber:
				idx -= 1
				pic = l[idx]
				if (pic.publish and (pic.sizex >= size or pic.sizey >= size)) or category == None:
					maxnumber -= 1
					liste.append(self._pictureAsHash(pic))
			return liste
		finally: self._release()
	
	# this method returns a hash with data from the selected image
	def getPicture(self, id):
		try:
			self._acquire()
			(idx, found) = self.pictures.locate({'id':int(id)})
			if found:
				pic = self.pictures[idx]
				return self._pictureAsHash(pic)
			else:
				return {}
		finally: self._release()
	
	# this method returns a hash with data from the image before the
	# selected image
	def getPreviousPicture(self, id, size=None):
		try:
			self._acquire()
			(idx, found) = self.pictures.locate({'id':int(id)})
			if found and idx:
				idx -= 1
				sizex = 0
				sizey = 0
				if size:
					fmt = self.getFormat(size)
					sizex = fmt['sizex']
					sizey = fmt['sizey']
				while idx >= 0 and (not(self.pictures[idx].publish) or (self.pictures[idx].sizex < sizex and self.pictures[idx].sizey < sizey)):
					idx -= 1
				if idx >= 0:
					pic = self.pictures[idx]
					return self._pictureAsHash(pic)
				else:
					return {}
			else:
				return {}
		finally: self._release()
	
	# this method returns a hash with data from the image after the
	# selected image
	def getNextPicture(self, id, size=None):
		try:
			self._acquire()
			(idx, found) = self.pictures.locate({'id':int(id)})
			if found:
				idx += 1
				sizex = 0
				sizey = 0
				maxidx = len(self.pictures)
				if size:
					fmt = self.getFormat(size)
					sizex = fmt['sizex']
					sizey = fmt['sizey']
				while idx < maxidx and (not(self.pictures[idx].publish) or (self.pictures[idx].sizex < sizex and self.pictures[idx].sizey < sizey)):
					idx += 1
				if idx < maxidx:
					pic = self.pictures[idx]
					return self._pictureAsHash(pic)
				else:
					return {}
			else:
				return {}
		finally: self._release()

	# this method returns a list of days where there are pictures for
	# the given size
	def getItemDays(self, category):
		try:
			self._acquire()
			sizex = 0
			sizey = 0
			if category:
				res = self.formats.select({'name':category})
				if len(res):
					sizex = res[0].sizex
					sizey = res[0].sizey
			days = {}
			res = []
			for pic in self.pictures:
				if (pic.publish and (pic.sizex >= sizex or pic.sizey >= sizey)) or category == None:
					(year, month, day) = (time.localtime(pic.pubtime))[:3]
					k = '%d-%d-%d' % (year, month, day)
					if not(days.has_key(k)):
						days[k] = 1
						res.append((year, month, day))
			return res
		finally: self._release()
	
	# this method checks wether a special day has pictures for a given size
	def hasItemsForDay(self, year, month, day, category):
		try:
			self._acquire()
			macros = self.getToolByNamespace('macros')
			sizex = 0
			sizey = 0
			if category:
				res = self.formats.select({'name':category})
				if len(res):
					sizex = res[0].sizex
					sizey = res[0].sizey
			(min, max) = macros.getMinMaxEpoch(year, month, day)
			for pic in self.picturesByPubtime.select({'pubtime':min},{'pubtime':max}):
				if (pic.publish and (pic.sizex >= sizex or pic.sizey >= sizey)) or category == None:
					return 1
			return 0
		finally: self._release()

	# this method returns the page for a calendar selection. This
	# forces all calendar clicks on the desktop to go to index_html
	def getCalendarUrl(self, year=None, month=None, day=None, category=None):
		if category:
			if day:
				return '%s;%04d-%02d-%02d?category=%s' % (self.getUrl(), year, month, day, category)
			else:
				return '%s;%04d-%02d?category=%s' % (self.getUrl(), year, month, category)
		else:
			if day:
				return '%s;%04d-%02d-%02d' % (self.getUrl(), year, month, day)
			else:
				return '%s;%04d-%02d' % (self.getUrl(), year, month)
	
	# this method returns the title of a page
	def getPageTitle(self, context):
		try:
			self._acquire()
			title = PyDS.Tool.StandardTool.getPageTitle(self, context)
			if self.prefs[0].title:
				title = self.prefs[0].title
			if context.get('pictureid',''):
				pic = self.getPicture(int(context['pictureid']))
				if pic:
					title = pic['title']
			return title
		finally: self._release()
		
	# this method returns the description of a page
	def getPageDescription(self, context):
		try:
			self._acquire()
			desc = PyDS.Tool.StandardTool.getPageDescription(self, context)
			if self.prefs[0].desc:
				desc = self.prefs[0].desc
			if context.get('pictureid',''):
				pic = self.getPicture(int(context['pictureid']))
				if pic:
					desc = pic['desc']
			return desc
		finally: self._release()
		
	# this method returns a list of format definitions
	def getFormatList(self, all=0):
		try:
			self._acquire()
			liste = []
			for format in self.formats:
				if format.render or all:
					liste.append({
						'id': format.id,
						'name': format.name,
						'sizex': format.sizex,
						'sizey': format.sizey,
						'render': format.render
					})
			return liste
		finally: self._release()
	
	# this method returns a named format
	def getFormat(self, name):
		try:
			self._acquire()
			res = self.formats.select({'name':name})
			if len(res):
				format = res[0]
				return {
					'id': format.id,
					'name': format.name,
					'sizex': format.sizex,
					'sizey': format.sizey,
					'render': format.render
				}
			else:
				return {}
		finally: self._release()
	
	# this method returns the filename in the PictureDatabase
	def pictureName(self, id):
		return os.path.join(self.picturedb, '%d.png' % id)
	
	# this method returns the filename of the thumbnail in the database
	def pictureThumbName(self, id):
		return os.path.join(self.picturedb, '%d-thumb.jpeg' % id)
	
	# this method removes all rendered output for a given picture
	def unrenderPicture(self, id):
		mirror = self.getToolByNamespace('mirror')
		pic = self.getPicture(id)
		self.deleteFile('%d.html')
		self.deleteFile('%d.jpeg')
		self.deleteFile('%d.png')
		for format in self.getFormatList():
			self.deleteFile('%d.html', category=format['name'])
			self.deleteFile('%d.jpeg', category=format['name'])
			self.deleteFile('%d.png', category=format['name'])
		mirror.deleteItem(self, str(id))
	
	# this method renders a picture. This builds the picture image
	# itself and all related index pages
	def renderPicture(self, id):
		self.queue.append(self._renderPicture, id, None)
		picids = {}
		for format in self.getFormatList():
			np = self.getNextPicture(id, format['name'])
			if np:
				picids[np['id']] = 1
			pp = self.getPreviousPicture(id, format['name'])
			if pp:
				picids[pp['id']] = 1
		for id in picids.keys():
			self.queue.append(self._renderPicture, id, None)
		pic = self.getPicture(id)
		self.renderPictureIndexTimeframe(pic['pubtime'])
	
	# this method renders a picture online without using the queue
	def renderPictureOnline(self, id):
		mirror = self.getToolByNamespace('mirror')
		pic = self.getPicture(id)
		if pic['publish']:
			for format in self.getFormatList():
				if pic['sizex'] >= format['sizex'] or pic['sizey'] >= format['sizey']:
					req = self._getDummyRequest(now=pic['pubtime'])
					html = req.renderText(
						pic['text'],
						picture=pic,
						size=format['name'],
						section_level=2
					)
					src = req.renderPage(
						'BaseRendering',
						category='',
						pictureid=pic['id'],
						body=html
					)
					self.storeHTML(
						src,
						'%d.html' % pic['id'],
						category=format['name']
					)
			mirror.syncItem(self, str(id))
				
	
	# this method renders the index pages for a given timeframe. This
	# usually is called after a delete of a picture
	def renderPictureIndexTimeframe(self, reftime):
		(year, month, day) = (time.localtime(reftime))[:3]
		self.renderPictureIndex(year, month, day)
		self.renderPictureIndex(year, month, 0)
		yp = year
		mp = month
		mp -= 1
		if mp == 0:
			mp = 12
			yp -= 1
		self.renderPictureIndex(yp, mp, 0)
		self.renderPictureIndex(0, 0, 0)
	
	# this method renders a picture index for a given date. It can be
	# either all 0 for the main index, year/month for the month page
	# or year/month/day for a day page
	def renderPictureIndex(self, year, month, day=0):
		self.queue.append(self._renderPictureIndex, (year, month, day), None)
	
	# this method renders a picture index online, without using the queue
	def renderPictureIndexOnline(self, year, month, day):
		req = self._getDummyRequest(datespec=(year, month, day))
		html = req.renderPage('PictureRendering',
			year=year, month=month, day=day, category=''
		)
		src = req.renderPage('BaseRendering', body=html, category='')
		self.storeHTML(src, year=year, month=month, day=day)
		if not(year) and not(month) and not(day):
			xml = req.renderPage('RSSFeedRendering', category='')
			self.storeXML(xml, category='')
	
	# this method returns an URL for the day of a picture
	def getCloudUrlForDay(self, reftime):
		(year, month, day) = time.localtime(reftime)[:3]
		return self.getStaticUrl(year=year, month=month, day=day)
	
	# this method delivers a menu bar with image sizes for a picture
	def imageSizeBar(self, id, form=''):
		pic = self.getPicture(id)
		liste = []
		for format in self.getFormatList():
			if pic['sizex'] >= format['sizex'] or pic['sizey'] >= format['sizey']:
				if form != format['name']:
					liste.append(
						'<a href="%s">%s</a>' % (
							self.getStaticUrl(
								'%d.html' % pic['id'],
								category=format['name']
							),
							format['name']
						)
					)
				else:
					liste.append('<b>%s</b>' % format['name'])
		return string.join(liste, ' | ')

	# this method returns an image url for a picture and resizes the
	# picture to the given boundingbox. The size can be given in the
	# same way as with imageTag.
	def imageUrl(self, id, size=None, picformat=None, interpolation=None):
		pic = self.getPicture(id)
		if picformat and pictureFormats.has_key(picformat):
			pic['format'] = picformat
		if interpolation:
			pic['interpolation'] = interpolation
		sizex = pic['sizex']
		sizey = pic['sizey']
		ext = pictureFormats[pic['format']][2]
		picname = '%d.%s' % (id, ext)
		category = ''
		if type(size) == type((1,)):
			(sizex, sizey) = size
			picname = '%d-%d-%d.%s' % (id, sizex, sizey, ext)
		elif type(size) == type(''):
			format = self.getFormat(size)
			if format:
				category = format['name']
				sizex = format['sizex']
				sizey = format['sizey']
		if sizex and sizey:
			ipath = self.pictureName(id)
			opath = self.getFileName(picname, category)
			if os.path.exists(opath):
				istat = os.stat(ipath)
				ostat = os.stat(opath)
				if istat[stat.ST_MTIME] >= ostat[stat.ST_MTIME] or pic['timestamp'] >= ostat[stat.ST_MTIME]:
					img = self._thumbnail(ipath, sizex, sizey, format=pic['format'], interpolation=pic['interpolation'])
					self.storeFile(img, picname, category)
			else:
				img = self._thumbnail(ipath, sizex, sizey, format=pic['format'], interpolation=pic['interpolation'])
				self.storeFile(img, picname, category)
			return self.getStaticUrl(
				picname,
				category=category,
			)
		return ''

	# this method returns an image tag for a picture and resizes the
	# picture to the given boundingbox. If the bounding box is missing,
	# the actual picture size is used. The format can be overridden
	# or is taken from the picture itself. If bigpage is set and url
	# is not, the url is constructed as a link to the big image page.
	# bigpage should be set to a format name.
	def imageTag(self, id, size=None, align="left", url=None, picformat=None, bigpage=None, interpolation=None):
		pic = self.getPicture(id)
		if picformat and pictureFormats.has_key(picformat):
			pic['format'] = picformat
		if interpolation:
			pic['interpolation'] = interpolation
		sizex = pic['sizex']
		sizey = pic['sizey']
		ext = pictureFormats[pic['format']][2]
		picname = '%d.%s' % (id, ext)
		category = ''
		if type(size) == type((1,)):
			(sizex, sizey) = size
			picname = '%d-%d-%d.%s' % (id, sizex, sizey, ext)
		elif type(size) == type(''):
			format = self.getFormat(size)
			if format:
				category = format['name']
				sizex = format['sizex']
				sizey = format['sizey']
		if sizex and sizey:
			ipath = self.pictureName(id)
			opath = self.getFileName(picname, category)
			if os.path.exists(opath):
				istat = os.stat(ipath)
				ostat = os.stat(opath)
				if istat[stat.ST_MTIME] >= ostat[stat.ST_MTIME] or pic['timestamp'] >= ostat[stat.ST_MTIME]:
					img = self._thumbnail(ipath, sizex, sizey, format=pic['format'], interpolation=pic['interpolation'])
					self.storeFile(img, picname, category)
			else:
				img = self._thumbnail(ipath, sizex, sizey, format=pic['format'], interpolation=pic['interpolation'])
				self.storeFile(img, picname, category)
			resized = Image.open(opath).size
			html = '<img src="%s" alt="%s" border=0 align="%s" width="%s" height="%s" hspace=3 vspace=3>' % (
				self.getStaticUrl(
					picname,
					category=category,
				),
				pic['title'],
				align,
				resized[0],
				resized[1]
			)
			if bigpage:
				format = self.getFormat(bigpage)
				if not(format):
					bigpage = ''
			if not(url) and bigpage and pic['publish']:
				url = self.getStaticUrl(
					'%d.html' % id,
					category=bigpage
				)
			if url:
				html = '<a href="%s" title="%s">%s</a>' % (
					url,
					pic['title'],
					html
				)
			return html
		return ''
	
	# this method returns the title of the RSS feed
	def getRSSTitle(self, category=''):
		macros = self.getToolByNamespace('macros')
		return macros.quoteXML(self.getPageTitle(context={}))
	
	# this method returns the description of the RSS feed
	def getRSSDescription(self, category=''):
		macros = self.getToolByNamespace('macros')
		return macros.quoteXML(self.getPageDescription(context={}))
	
	def _translateItemToRSS(self, pic, format):
		macros = self.getToolByNamespace('macros')
		item = {
			'title': macros.quoteXML(pic['title']),
			'link': self.getStaticUrl(page='%d.html' % pic['id'], category=format),
			'guid': self.getStaticUrl(page='%d.html' % pic['id'], category=format),
			'pubdate': macros.TodayFull(now=pic['pubtime']),
			'source': '',
			'sourceurl': '',
			'categories': [],
			'comments': self.weblogCommentsUrl(pic['id'], pic['pubtime']),
			'trackback': self.weblogTrackbackUrl(pic['id'], pic['pubtime']),
			'description': macros.quoteXML(
				'%s %s' % (
					self.imageTag(pic['id'], (150, 150)),
					pic['desc']
				)
			)
		}
		return item
	
	# this method returns one picture as a RSS item.
	# fullrss is ignored, descriptions are allways full.
	def getRSSItem(self, id, category='', fullrss=None):
		format = self.getFormatList()[0]['name']
		pic = self.getPicture(id)
		return self._translateItemToRSS(pic, format)
	
	# this method returns the pictures as an item list, orderd by time
	def getRSSItems(self, category=''):
		items = []
		format = self.getFormatList()[0]['name']
		for pic in self.getRecentPictures():
			if pic['publish']:
				item = self._translateItemToRSS(pic, format)
				items.append(item)
		return items
	
	# this method returns hours without pictures
	def getRSSSkipHours(self, category=''):
		hours = {}
		for pic in self.getRecentPictures(maxnumber=100):
			hours[time.gmtime(pic['pubtime'])[3]] = 1
		nonhours = []
		for hour in range(24):
			if not(hours.has_key(hour)):
				nonhours.append(hour)
		return nonhours

	# This method gives a link tag that uses a thumbnail as link content.
	def getRSSItemLinkTag(self, item, postid, align='left'):
		return self.imageTag(
			int(postid),
			size=(100,100),
			align=align,
			url=item['guid']
		)

	# This methods gives you an cloud URL for a day. You can pass in the
	# reference time (as float).
	def getCloudUrlForDay(self, reftime):
		(year, month, day) = time.localtime(reftime)[:3]
		return self.getStaticUrl(year=year, month=month, day=day)
					
	# This methods gives you an cloud URL for a picture. You can pass in the
	# reference time (as float) and an item id. The picture is referenced
	# as a fragment of a day page, so you pass in a time value and the ID
	# of the posting.
	def getCloudUrlForItem(self, reftime, id):
		url = self.getCloudUrlForDay(reftime)
		url += '#'
		url += self.weblogAnchor(id)
		return url
	
	# this macro returns the weblog anchor name for a picture
	def weblogAnchor(self, id):
		return 'pic%d' % id
		
	# This macro creates a permalink for a picture posting. This is
	# a complete HTML a-tag.
	def weblogPermaLink(self, id, reftime, linktext='#'):
		html = '<a href="%s" title="%s">%s</a>' % (
			self.getCloudUrlForItem(reftime, id),
			_("Permalink to this posting"),
			linktext
		)
		return html
	
	# This macro returns the URL for comments (used in RSS)
	def weblogCommentsUrl(self, id, reftime):
		upstream = self.getToolByNamespace('upstream')
		macros = self.getToolByNamespace('macros')
		url = '%s?u=%s&amp;p=%s&amp;link=%s' % (
			upstream.get('prefs','commentsurl'),
			upstream.get('prefs','usernum'),
			self.weblogAnchor(id),
			macros.quoteURL(self.getCloudUrlForItem(reftime, id))
		)
		return url
	
	# This macro returns the comments text and link for postings
	def weblogCommentsLink(self, id, reftime):
		macros = self.getToolByNamespace('macros')
		src = '<a href="%s"'
		src += ' %s title="%s">%s'
		src += ' [<script type="text/javascript" '
		src += 'language="JavaScript">commentCounter(\'%s\')</script>]</a>'
		html = src % (
			self.weblogCommentsUrl(id, reftime),
			macros.onClick('comments'),
			_('Click here to comment this picture'),
			_('comment'),
			self.weblogAnchor(id)
		)
		return html
		
	# This macro returns the URL for trackbacks (used in RSS)
	def weblogTrackbackUrl(self, id, reftime):
		upstream = self.getToolByNamespace('upstream')
		macros = self.getToolByNamespace('macros')
		url = '%s?u=%s&amp;p=%s' % (
			upstream.get('prefs','trackbackurl'),
			upstream.get('prefs','usernum'),
			self.weblogAnchor(id)
		)
		return url
	
	# This macro returns the trackback text and link for postings
	def weblogTrackbackLink(self, id, reftime):
		macros = self.getToolByNamespace('macros')
		src = '<a href="%s"'
		src += ' %s title="%s">%s'
		src += ' [<script type="text/javascript" '
		src += 'language="JavaScript">trackbackCounter(\'%s\')</script>]</a>'
		html = src % (
			self.weblogTrackbackUrl(id, reftime),
			macros.onClick('trackbacks'),
			_('Click here to see the trackbacks for this picture'),
			_('trackback'),
			self.weblogAnchor(id)
		)
		return html
		
__desc__ = _("This is a tool that manages your Pictures.")

PyDS.Tool.registerTool(PictureTool, 'pictures', _('Pictures'), __desc__, 70)

