"""

Python Desktop Server - Structured Text support

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>
Copyright (c) 2003, Garth T. Kidd <garth@deadlybloodyserious.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: StructuredText.py,v 1.38 2004/07/05 17:35:05 gb Exp $

import docutils.nodes
import docutils.parsers.rst
import docutils.utils
import docutils.core
import docutils.io
import docutils.writers.html4css1
import docutils.readers.standalone
import docutils.transforms
import sys
import re
import urllib
import htmlentitydefs

import PyDS.docpytils

# there are problems with some systems with installing SilverCity. So
# try to import it and if it doesn't work, don't import codeblock!
try:
	import SilverCity
	import PyDS.codeblock
except: pass

try:
	from cStringIO import StringIO
except ImportError:
	from StringIO import StringIO

# a bad hack to preserver my installed translation handler, as docutils has
# it's own. Only patch this if we are loaded by PyDS, so ignore NameError
try:
	_PyDS.translation.install()
	runningUnderPyDS = 1
except NameError, e:
	runningUnderPyDS = 0

# if we are running under PyDS, pull in additional directives
if runningUnderPyDS: import PyDS.Directives

reunicode = re.compile(ur'[\u0080-\uffff]')
def entity(c):
	return '&#'+str(ord(c[0]))+';'

class HTMLFragmentTranslator(docutils.writers.html4css1.HTMLTranslator):
	"""Adaptation of the docutils `HTMLTranslator` to issue fragments
	rather than whole documents."""

	def __init__(self, document, section_level=1, suppress_paragraphs=0):
		docutils.writers.html4css1.HTMLTranslator.__init__(self, document)
		if section_level is None: # just in case
			section_level = 1
		self.section_level = section_level
		self.suppress_paragraphs = suppress_paragraphs

	def astext(self):
		"""Issue just the body, trimming obsolete div."""
		if len(self.body)>=2:
			if self.body[0].startswith('<div') \
				and self.body[-1].startswith('</div'):
				return ''.join(self.body[1:-1])
		return ''.join(self.body)

	def visit_paragraph(self, node):
		"""Issue a paragraph, ignoring `HTMLTranslator`'s usual habit of
		eliminating paragraph tags if there's only one paragraph in this
		parent. But only do this if suppress_paragraphs is 0."""
		if self.suppress_paragraphs:
			docutils.writers.html4css1.HTMLTranslator.visit_paragraph(self, node)
		else:
			self.body.append(self.starttag(node, 'p', ''))
			self.context.append('</p>\n')

	def visit_admonition(self, node, name):
		"""Override the default admonition renderer to put in
		fancy icons instead of text labels"""
		image = '<img src="/images/%s.gif" alt="%s">' % (name, name)
		if runningUnderPyDS:
			macros = sys.modules['PyDS.Tool'].getTool('macros')

			image = macros.imageTag('/images/%s.gif' % name, alt=name)
		self.body.append(self.starttag(node, 'div', CLASS=name))
		self.body.append('<div class="admonition-title">'
			+ image
			+ '</div>\n')
	def visit_field_name(self, node):
		atts = {}
		if self.in_docinfo:
			atts['class'] = 'docinfo-name'
		else:
			atts['class'] = 'field-name'
		if len(node.astext()) > 30:
			atts['colspan'] = 2
			self.context.append('</tr>\n<tr><td>&nbsp;</td>')
		else:
			self.context.append('')
		self.body.append(self.starttag(node, 'th', '', **atts))


class HTMLFragmentWriter(docutils.writers.html4css1.Writer):
	"""HTML `Writer` that issues fragments rather than whole documents."""

	def __init__(self, section_level=1, suppress_paragraphs=0):
		docutils.writers.html4css1.Writer.__init__(self)
		self.translator_class = HTMLFragmentTranslator
		self.section_level = section_level
		self.suppress_paragraphs = suppress_paragraphs
	
	def translate(self):
		visitor = self.translator_class(self.document,
			section_level=self.section_level,
			suppress_paragraphs=self.suppress_paragraphs
		)
		self.document.walkabout(visitor)
		self.visitor = visitor
		self.output = visitor.astext()
		self.head_prefix = visitor.head_prefix
		self.head = visitor.head
		self.body_prefix = visitor.body_prefix
		self.body = visitor.body
		self.body_suffix = visitor.body_suffix

_Transforms = []

def registerTransform(transformClass): 
	"Register a transform class with the renderer."
	_Transforms.append(transformClass)

from docutils.readers.standalone import Reader as standaloneReader
class Reader(standaloneReader):
	"""A customised `Reader` that applies our custom transforms."""
	# supported = standaloneReader.supported + ('wiki',) # why?
	def __init__(self, *args, **kw): 
		res = standaloneReader.__init__(self, *args, **kw)
		self.default_transforms = tuple(_Transforms) + standaloneReader.default_transforms
		return res

class Latin1StringInput(docutils.io.StringInput):
	"""disable the unicode decoding of structured text"""
	def decode(self, data):
		return data
	
	def read(self, dummy=None):
		return docutils.io.StringInput.read(self)

class Latin1StringOutput(docutils.io.StringOutput):
	"""disable the unicode decoding of structured text"""
	def write(self, data):
		if type(data) == type(u''):
			data = reunicode.sub(lambda m: entity(m.group(0)), data).encode(_PyDS.documentEncoding)
		self.destination = data
		return self.destination

def renderText(text, section_level=1, suppress_paragraphs=0, uniqueId=None, homeTool=None):
	"""Render `text`'s contents to XHTML.
	
	uniqueId -- unique ID of the item being rendered
	homeTool -- the home tool's instance"""
	
	try: 
		# make uniqueId and homeTool available to the renderer, so 
		# that the MeshTool (if registered) can determine backlinks. 
		if runningUnderPyDS: 
			_flet.begin(uniqueId=uniqueId, homeTool=homeTool)
		# crank it through
		writer = HTMLFragmentWriter(
			section_level=section_level,
			suppress_paragraphs=suppress_paragraphs)
		settings = { 'output_encoding': 'unicode' }
		if runningUnderPyDS: 
			settings['input_encoding'] = _PyDS.documentEncoding
		output = docutils.core.publish_string(
			source=text,
			reader=Reader(),
			writer=writer,
			parser_name='restructuredText',
			settings_overrides=settings
		)
		try:
			if runningUnderPyDS: 
				args = [_PyDS.documentEncoding]
			else:
				args = []
			output = output.encode(*args)
		except:
			output = re.sub(
				r'([^\x00-\xff])',
				lambda m: '&#%d;' % ord(m.group(1)),
				output
			)
			output = output.encode(_PyDS.documentEncoding)
	finally:
		if runningUnderPyDS:
			_flet.end()

	return output

	return destination.destination

if __name__ == '__main__':
	import time
	try: 
		import psyco
		psyco.log()
		psyco.profile()
		print "PSYCO ENABLED."
	except ImportError: 
		pass
	testText = """
This is a headline
------------------

This is a paragraph.

This is another paragraph. 

1. first item
2. second item
3. third item

And an external Link_.  And an internal `Link 3`_. And another internal `Link 4`_.

.. This is a comment.

.. Comment: this is a comment, too.

.. DANGER:: Inset block to warn about something.
.. WARNING:: Inset block to warn about something.

.. note:: everthing inset under this note will be included.

   - first line of list
   - second line of list

   This is another text of the note.

   .. image:: http://muensterland.org/images/muensterland.png
      :width: 100
      :height: 100
      :alt: muensterland logo

And a [CIT001]_ citation. And a [2]_ footnote.

And an anonymous__ link. And another__ one.

.. _Link 3:

This is a referenced paragraph by _`Link 4`.

.. _Link: http://pyds.muensterland.org/

.. [CIT001] This is a citation.

.. [2] This is a footnote

.. __: http://www.f-2.org/
.. __: http://muensterland.org/

"""
	LOOPS=50
	begin = time.time()
	print "Rendering example text %d times" % LOOPS,
	for i in range(LOOPS): 
		renderedText = renderText(testText)
		print ".", 
	print
	elapsed = time.time() - begin
	print renderedText
	print "Rendered example text %d times in %f seconds (%f renders/s)" % (
			LOOPS,
			elapsed,
			LOOPS/elapsed)

