"""

Python Desktop Server - Wiki Editor

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>
Copyright (c) 2002, Garth T Kidd <garth@deadlybloodyserious.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: WikiTool.py,v 1.27 2004/08/04 10:56:36 gb Exp $

import PyDS.Tool
import metakit
import time
import string
import sys
import cgi
import re
import os.path
    
def _munge(s): 
	"Elminiate non-alphanumeric and make lowercase."
	t = re.sub('\W', '', s)
	return t.lower()

def _splitPath(path): 
	"Split a path to a tuple, with '' as the first item"
	p, b = os.path.split(path)
	if not p: 
		return (b,)
	else: 
		return _splitPath(p) + (b,)

def _normPath(path): 
	"Normalise a path a little -- primarily, converting \\ to /."
	return '/'.join(_splitPath(path))

class WikiTool(PyDS.Tool.StandardTool):

	def _createdb(klass, db):
		db.getas("nodes[id:S,title:S,text:S,pubdate:S,pubtime:F,categoryid:S,baseid:S,template:I,cloud:S]")
		db.getas("categories[id:S]")

	_createdb = classmethod(_createdb)

	def _initdb(self):
		try:
			self._acquire()
			self.nodes = self.db.view('nodes').ordered(1)
			self.categories = self.db.view('categories')
			self._commit()
		finally: self._release()

	
	def _backup(self):
		l = self._backupTable('nodes')
		return l
			
	def _initopts(self):
		self.rss = 1
		self.showCalendar =0 
		if self.level>0 and self._getNode('index') is None: 
			# If I add a preference for rendering Wiki pages to the main site 
			# rather than a tool-specific subdirectory, I'll need to suppress 
			# automatic creation of this file. 
			self.addNode('index', 
			             _("Wiki Index"), 
			             _("Edit this page to begin your Wiki"))
		if self._getNode('404') is None: 
			self.addNode('404', 
			             _("Unresolved Target"), 
			             _("Not-yet-resolved links in the Wiki are pointed here."))
	
	def _initactive(self):
		navigation = self.getToolByNamespace('navigation')
		themes = self.getToolByNamespace('themes')
		navigation._addHook('select', 'wiki',
			_('Link to a wiki page'),
			self._wikiNavHook)
		themes.hookTemplate(
			'WikiRendering',
			_('wiki template for the cloud'),
			'templates/WikiRendering.tmpl'
		)
		
	def _wikiNavHook(self, qual, value, preset=None):
		if qual == 'html':
			html = '<select name="%s">' % value
			for node in self.getNodes():
				if not(node['template']):
					selected = ''
					if node['id'] == preset:
						selected = ' selected'
					html += '<option value="%s"%s>%s' % (
						node['id'],
						selected,
						node['title']
					)
			html += '</select>'
			return html
		elif qual == 'url':
			return self.getCloudUrlForNode(value)
		else:
			raise ValueError('%s is not a valid qualifier' % qual)
	
	def _initthread(self, server):
		while server.isRunning():
			time.sleep(10)
			if self.queue.length():
				while self.queue.length():
					(jobf, key, data) = self.queue.shift()
					jobf(key, data)
				# since we had something to render, we should
				# now render the RSS feed to get a current one
				req = self._getDummyRequest()
				xml = req.renderPage('RSSFeedRendering', category='')
				self.storeXML(xml, category='')
				self.logVerbose(_('work done, going to sleep again'))
	
	def _status(self):
		html = PyDS.Tool.StandardTool._status(self)
		html += '<ul>'
		html += _('<li>There are %d wiki nodes in the database</li>') % len(self.nodes)
		html += _('<li>There are %d items in the render queue</li>') % self.queue.length()
		html += self._threadstatus()
		html += '</ul>'
		return html
	
	def _search(self, query):
		return [] # not implemented yet

	def _resync(self, theMirror):
		for node in self.nodes:
			if not(node.template):
				theMirror.iHave(self, node.id)

	# this method returns an URL to the cloud in the context of the current tool
	# or the given tool. Weblog patches the URL if radio compatibility is requested.
	def getStaticUrl(self, page=None, tool=None, year=None, month=None, day=None, category='', base=None, flat=0):
		url = PyDS.Tool.StandardTool.getStaticUrl(self, page, tool, year, 
		                                          month, day, category, 
												  base, flat, stripToolPrefix=1)
		# PyDS.Tool.debugOutput("=> %s" % url)
		return url

	# This method returns the URL for a node. Depending on context it
	# returns a desktop URL or a cloud URL.
	def getCloudUrlForNode(self, id): 
		if _flet.desktop:
			return self.getUrl('show_html', id=str(id))
		else:
			node = self.getNode(id)
			if node and not(node['template']):
				return self.getStaticUrl(page='%s.html' % id)
			else:
				return self.getStaticUrl(page='404.html')
		
	# --------------------------------------------------------------------
	# NEW STRAY LINK RESOLUTION METHODS
	# --------------------------------------------------------------------

	# Find items potentially matching the link text, and return their unique IDs. 
	def getUniqueIdsMatchingLinkText(self, linkText): 
		"""Find items potentially matching `linkText`. 
		Return a list of string IDs uniquely identifying those items."""
		if not linkText: 
			return # never match empty link text

		# precalculate more permissive versions of the link text
		m = _munge(linkText)
		lt = linkText.lower()

		# precalculate permissive versions of context-dependent variables
		conid = _flet.get('uniqueId', '').lower() # uniqueId
		concatid = _flet.get('categoryid', '').lower() # categoryid
		
		# define filters, each of which depends on local variables defined 
		# previously -- a lot easier than passing lots of arguments, no?
		orderedFilters = (
			# id match to link text
			lambda: lid == lt, 
			# id match on munged text
			lambda: lid == m, 
			# subnode of current (according to context) node
			lambda: conid and catid == conid and baseid == m, 
			# peer node of current (according to context) node
			lambda: concatid and catid == concatid and baseid == m, 
			# baseid match
			lambda: baseid == m, 
			# title match
			lambda: m == _munge(node.title),
			)

		# prepare filter results dictionary
		filterResults = {}
		for filter in orderedFilters: 
			filterResults[filter] = []

		# run each node through the filters
		# if CPU becomes a bigger worry than memory, we can cache all these
		try: 
			self._acquire()
			for node in self.nodes:
				if not(node.template):
					# precalculate more permissive version of node attributes
					nid = node.id
					lid = nid.lower()
					catid = node.categoryid.lower()
					baseid = node.baseid.lower()
					# execute the filters
					for filter in orderedFilters: 
						try: 
							if filter(): 
								filterResults[filter].append(nid)
						except: 
							pass # not worried about filter failure
		finally:
			self._release()

		# gather results into a single list, preserving order
		allResults = []
		inAllResults = {}
		for filter in orderedFilters: 
			for result in filterResults[filter]: 
				if inAllResults.has_key(result): 
					continue
				else: 
					inAllResults[result] = 1
					allResults.append(result)
		return allResults

	# Return the upload target for a given item in the current context. 
	def getCloudLinkDetailsForUniqueId(self, uniqueId): 
		"""Return the upload target for `uniqueId` in the current context.
		Raise KeyError if `uniqueId` cannot be found."""
		node = self.getNode(uniqueId)
		if node and not(node['template']):
			return ('wiki', self.getStaticUrl(page='%s.html' % uniqueId), node.get('title',''))
		else: 
			raise KeyError, uniqueId

	# Return the upload target for a nonexistent page for the current context. 
	def getCloudLinkDetailsFor404(self):
		cssclass, url, title = self.getCloudLinkDetailsForUniqueId('404')
		return 'undefinednode', url, title

	# Return a URL for an edit page for a given item.
	def getDesktopDisplayLinkDetailsForUniqueId(self, uniqueId): 
		"""Return the edit URL for `uniqueId`."""
		node = self.getNode(uniqueId)
		if node and not(node['template']):
			return ('wiki', self.getUrl('show_html', id=uniqueId), node.get('title',''))
		else: 
			raise KeyError, uniqueId

	# Return a link to a creation page for an item, based on link text. 
	def getDesktopCreateLinkDetailsForProposedLinkText(self, linkText): 
		"""Return the edit URL for a non-existent item, based on `linkText`."""
		# split linkText into words on whitespace
		words = re.split(r"\s+", str(linkText))
		# remove non-alphanumerics from each word
		words = [re.sub(r"\W", '', word) for word in words]
		# capitalise each word
		words = [word[0:1].upper() + word[1:].lower() for word in words]
		# string 'em together
		id = ''.join(words)
		# figure out our Wiki context, and prepend it if found
		concatid = _flet.get('categoryid', '').lower()
		if concatid: 
			id = '/'.join(concatid, id)
		# et voila! 
		return ('undefinednode', 
		        self.getUrl('edit_html', id=id),
		        _("Create this item in WikiTool"))

	# Render an item by its unique ID
	def renderItemByUniqueId(self, uniqueId):
		return self.renderNode(uniqueId)

	# --------------------------------------------------------------------
		
	# this method returns the title of a page for rendering. This does a
	# lookup of data in the category or in the preferences and adds
	# date specifiers. The context passed in is the context that you
	# get in your template.
	def getPageTitle(self, context):
		if context.has_key('title'): 
			return context['title']
		log('WikiTool.getPageTitle failed to find title; context.keys==[%s]', 
		    ', '.join(map(repr, context.keys())))
		title = PyDS.Tool.StandardTool.getPageTitle(self, context)
		if context.get('category',''):
			cat = self.getCategory(context['category'])
			if cat:
				title = cat['title']
		if context.get('year',0) and context.get('month',0) and context.get('day',0):
			title += ' %d.%d.%d' % (context['day'], context['month'], context['year'])
		elif context.get('year',0) and context.get('month',0):
			title += ' %d/%d' % (context['year'], context['month'])
		return title
		
	# this method returns the description of a page for rendering. It's
	# a sister to the getPageTitle method.
	def getPageDescription(self, context):
		desc = PyDS.Tool.StandardTool.getPageDescription(self, context)
		if context.get('category',''):
			cat = self.getCategory(context['category'])
			if cat:
				desc = cat['desc']
		return desc
		
	def index_html(self, req):
		macros = self.getToolByNamespace('macros')
		html = [_("<h3>Your Wiki Nodes</h3>")]
		append = html.append
		editvars = {
			'edit_html': self.getUrl('mux_redir'), 
			'editlabel': _("New")
		}
		append('<div id="toolbar"><form action="%(edit_html)s" method="POST">' % editvars)
		append('<input type="submit" name="WikiNew" value="%(editlabel)s" accesskey="n"></div>' % editvars)

		hashedNodes = {}
		for node in self.nodes: 
			nids = _splitPath(node.id)
			current = hashedNodes
			for id in nids: 
				idnode = current.setdefault(id, [None, {}])
				current = idnode[1]
			idnode[0] = node
		def listRenderer(hashedNodes): 
			keys = hashedNodes.keys()
			keys.sort()
			html.append('<ul>')
			for k in keys: 
				html.append('<li>')
				node, hashedChildren = hashedNodes[k]
				if node is not None: 
					nodevars = {
						'url': self.getUrl('show_html', id=str(node.id)),
						'title': node.title,
						'id': k,
						'lastmod': macros.Today(node.pubtime, withtime=1),
					}
					if node.template:
						html.append('<img src="%simages/new.gif" alt="%s">' % (
							self.pathprefix(),
							_('This node is a template node')
						))
					html.append('<a href="%(url)s" title="%(lastmod)s">%(id)s</a>: %(title)s' % nodevars)
				else: 
					html.append('%s' % k)
				if hashedChildren: 
					listRenderer(hashedChildren)
				html.append('</li>')
			html.append('</ul>')
		listRenderer(hashedNodes)
		html.append('</form>')
		req.setLocalValue('title', _('Your Wiki Nodes'))
		return req.renderPage('BaseTemplate', body='\n'.join(html))

	def show_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		node = self.getNode(id)
		subnodes = req.getFirstValue('subnodes')
		if 1: # subnodes and subnodes=='yes': 
			subnodehtml = self.renderSubNodes(id, depthLimit=1)
		else: 
			subnodehtml = ''
		template = ''
		if node.get('template',0):
			template = _(' (Template)')
		# used to be wrapped in 
		# _flet.begin(categoryid=node.get('categoryid',''), id=id)
		rendered = req.renderText(
			node.get('text',_('This node has no text.')), 
			structured = 2, # TODO: check 'structured' pref
			uniqueId = id, 
			homeTool = self) 
		viewhtml = ''
		if not(node.get('template',0)):
			viewhtml = '<input type="submit" name="WikiView" value="%(viewlabel)s" accesskey="v">' % {
				'viewlabel': _("View"), 
				'viewurl': self.getStaticUrl(page='%s.html' % id),
			}
		html = """\
%(breadcrumbs)s
%(subnodes)s
<div id="toolbar"><form action="%(mux_redir)s" method="POST">
<input type="hidden" name="id" value="%(id)s">
<input type="hidden" name="concatid" value="%(concatid)s">
<input type="submit" name="WikiNew" value="%(newlabel)s" accesskey="n">
<input type="submit" name="WikiEdit" value="%(editlabel)s" accesskey="e">
<input type="submit" name="WikiDelete" value="%(deletelabel)s">
%(viewhtml)s
<input type="submit" name="WikiClone" value="%(clonelabel)s" accesskey="c">
<input type="submit" name="WikiBlog" value="%(bloglabel)s" accesskey="b">
<!-- input type="submit" name="WikiSubNodes" value="%(subnodeslabel)s" accesskey="s" -->
</form></div>
<h2>%(title)s%(template)s</h2>
%(rendered)s
<p align="right" class="smallfont">%(lastlabel)s %(pubtimestr)s</p>
<p align="right"><a href="%(toolurl)s" accesskey="i">%(index)s</a></p>""" % {
			'title': cgi.escape(node.get('title',_('Set your page title here'))), 
			'breadcrumbs': self.renderBreadCrumbs(id, startWithTool=1), 
			'subnodes': subnodehtml, 
			'mux_redir': self.getUrl('mux_redir'),
			'id': str(id), 
			'concatid': '/'.join(_splitPath(str(id))[:-1]), 
			'newlabel': _("New"), 
			'editlabel': _("Edit"), 
			'deletelabel': _("Delete"), 
			'clonelabel': _("Clone"),
			'bloglabel': _("Blog"),
			'subnodeslabel': _("SubNodes"),
			'rendered': rendered, 
			'lastlabel': _("last change"), 
			'pubtimestr': macros.Today(node.get('pubtime',time.time()), withtime=1), 
			'toolurl': self.getUrl(), 
			'viewhtml': viewhtml,
			'index': _("Wiki index"),
			'template': template,
		}
		req.setLocalValue('title', node.get('title',_('Set your page title here')))
		return req.renderPage('BaseTemplate', body=html, title="Wiki: Show: %s" % node.get('title', '(no title)'))
		
	def mux_redir(self, req): 
		id = req.getFirstValue('id')
		if not id: 
			id = ''
		if req.getFirstValue('WikiNew'): 
			concatid = id # req.getFirstValue('concatid')
			if concatid: 
				id = concatid + '/<new node>'
			else: 
				id = '<new node>'
			return self.getUrl('edit_html', id=str(id), title='', text='')
		if req.getFirstValue('WikiEdit'): 
			return self.getUrl('edit_html', id=str(id))
		if req.getFirstValue('WikiClone'): 
			return self.getUrl('edit_html', id='', template=str(id))
		if req.getFirstValue('WikiDelete'): 
			return self.getUrl('delete_html', id=str(id))
		if req.getFirstValue('WikiView'): 
			return self.getStaticUrl(page='%s.html'%id)
		if req.getFirstValue('WikiBlog'): 
			node = self.getNode(id)
			return self.getUrl('index_html',
				tool='weblog',
				ptitle=node['title'],
				plink=self.getStaticUrl(page='%s.html' % id)
			)
		if req.getFirstValue('WikiSubNodes'): 
			return self.getUrl('show_html', id=str(id), subnodes='yes')
		return self.getUrl('show_html', id=str(id))

	def delete_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		node = self.getNode(id)
		html = """\
%(breadcrumbs)s
<div id="toolbar"><form action="%(delete_redir)s" method="POST">
<input type="hidden" name="id" value="%(id)s">
<input type="submit" name="WikiDeleteConfirm" value="%(confirmlabel)s" accesskey="d">
<input type="submit" name="WikiDeleteCancel" value="%(cancellabel)s" accesskey="c">
</form></div>
<h2>%(title)s</h2>
%(rendered)s
<p align="right" class="smallfont">%(lastlabel)s %(pubtimestr)s</p>""" % {
			'title': cgi.escape(node['title']), 
			'show_html': self.getUrl('show_html', id=str(id)), 
			'delete_redir': self.getUrl('delete_redir', id=str(id)), 
			'id': str(id), 
			'cancellabel': _("Cancel"), 
			'confirmlabel': _("Confirm Deletion"), 
			'rendered': req.renderText(
				node['text'], 
				structured = 2, # TODO: check 'structured' pref
				#categoryid=node['categoryid'], 
				#id=id,
				uniqueId = id, 
				homeTool = self),
			'lastlabel': _("last change"), 
			'pubtimestr': macros.Today(node['pubtime'], withtime=1), 
			'breadcrumbs': self.renderBreadCrumbs(id, startWithTool=1),
		}
		req.setLocalValue('title', node['title'])
		return req.renderPage('BaseTemplate', body=html, title="Wiki: Delete: %s" % node.get('title', '(no title)'))

	def edit_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		if not id:
			id = ''
		template = req.getFirstValue('template')
		if not template:
			template = ''
		node = self.getNode(id)
		if not(node) and not(template):
			templates = self.getTemplates()
			if templates:
				html = _('<p>Select your template for the new page:')
				html += '<form action="%s">' % req.getUrl('edit_html')
				html += '<input type="hidden" name="id" value="%s">' % id
				html += '<select name="template">'
				html += '<option value="404">%s' % _('Default: empty node')
				for template in templates:
					html += '<option value="%(id)s">%(title)s' % template
				html += '</select>'
				html += '<input type="submit" value="%s">' % _('Select')
				html += '<form>'
				req.setLocalValue('title', _('Select a template below'))
				return req.renderPage('BaseTemplate', body=html, title='Wiki: Edit: "%s"' % (id or _('new node')))
		elif not(node) and template:
			if template != '404':
				node['title'] = self.getNode(template)['title']
				node['text'] = self.getNode(template)['text']
		values = {} 
		values['idvalue'] = node.get('id', id)
		values['titlevalue'] = node.get('title', '')
		values['textvalue'] = node.get('text', '')
		for field in ['Id', 'Title', 'Text', 'delete', 'save']: 
			values[field.lower()+'label'] = _(field)
		values['edit_redir'] = self.getUrl('edit_redir')
		values['templatelabel'] = _('Template')
		values['templateselect'] = macros.checkBox('template', node.get('template',0))
		
		# Make a copy of the full ID before we truncate it
		values['oldidvalue'] = values['idvalue']
		
		# Try and figure out the context selector
		catids = []
		segs = _splitPath(id)
		values['idvalue'] = segs[-1]
		for seg in segs[:-1]: 
			if catids: 
				catids.append('%s/%s' % (catids[-1], seg))
			else: 
				catids.append(seg)
		catids.reverse()
		catids.append(_("(The Wiki)"))
		if node or len(catids)<2: 
			catidvalue = catids[0]
		else: 
			catidvalue = catids[1]
		values['breadcrumbs'] = self.renderBreadCrumbs(catids[0], startWithTool=1)
		catidoptions = []
		for catid in catids: 
			if catid == catidvalue: 
				selected = ' selected'
			else: 
				selected = ''
			catidoptions.append('<option value="%s"%s>%s</option>' % (
				cgi.escape(catid), selected, cgi.escape(catid)))
		values['catidoptions'] = '\n'.join(catidoptions)

		# Now for the text
		html = """
%(breadcrumbs)s
<form action="%(edit_redir)s" method="POST">
<table border=0>
<tr>
  <th align="right">Context</th>
  <td><select name="catid">%(catidoptions)s
</select></td>
</tr><tr>
  <th align="right">%(idlabel)s</th>
  <td><input name="oldid" value="%(oldidvalue)s" type="hidden">
      <input name="id" value="%(idvalue)s" size=50></td>
</tr><tr>
  <th align="right">%(titlelabel)s</th>
  <td><input name="title" value="%(titlevalue)s" size=50></td>
</tr><tr>
  <th align="right">%(templatelabel)s</th>
  <td>%(templateselect)s</td>
</tr><tr>
  <th valign="top" align="right">%(textlabel)s</th>
  <td><textarea nowrap name="text" rows=15 cols=60>%(textvalue)s</textarea></td>
</tr><tr>
  <td colspan=2 align="right">
    <input type=submit name="WikiSave" accesskey="s" value="%(savelabel)s">
  </td>
</tr></table></form>
""" % values
		html += '<p align="right"><a href="%s">%s</a>' % (
			self.getUrl(),
			_('choose a different node')
		)
		req.setLocalValue('title', _('Edit your node below'))
		return req.renderPage('BaseTemplate', body=html, title="Wiki: Edit: %s" % node.get('title', '(no title)'))

	def delete_redir(self, req):
		id = req.getFirstValue('id')
		if req.getFirstValue('WikiDeleteCancel'): 
			return self.getUrl('show_html', id=str(id))
		self.deleteNode(id)
		return self.getUrl()

	def invalid_id(self, id): 
		"Is the ID invalid? Why?"
		if ' ' in id: 
			return _("Spaces not allowed.")
		if re.match(r"^[\\\/].*", id): 
			return _("Can't start with a slash.")
		if re.match(r"^.*[\\\/]$", id): 
			return _("Can't end with a slash.")
		if not re.match(r"^\w+([\\\/]\w+)*$", id): 
			return _("Can't figure out exactly why. Probably a non-alphanumeric character.")
		splits = _splitPath(id)
		first = splits[0]
		try: 
			tool = self.getToolByNamespace(first.lower()) 
			return _("First path component %s conflicts with a tool name") % repr(first)
		except KeyError: 
			pass
		if re.match(r"^[0-9]{4}$", first): 
			return _("First path component %s might conflict with a year") % repr(first)
		return ''

	def fixid(self, id): 
		"Return a FixId for the invalid id."
		if id: 
			return '%s%s' % (
				   time.strftime("%Y%m%d%H%M_"), 
				   _munge(id))
		else:
			return ''
		
	def edit_redir(self, req):
		catid, id, oldid, title, text, desc, template = map(req.getFirstValue, 
			('catid', 'id', 'oldid', 'title', 'text', 'desc', 'template'))
		if req.getFirstValue('WikiDelete'):
			return self.getUrl('delete_html', id=str(id))
		else:
			if not id:
				id = ''
			errmsg = self.invalid_id(id)
			if errmsg: 
				id = self.fixid(id)
			# if we have a context, use it -- unless it starts with a 
			# bracket, in which case the context is the Wiki itself
			if catid and catid[0] != '(': 
				id = '/'.join((catid, id))
			if not text: 
				if id:
					return self.getUrl('edit_html', id=str(id), errmsg=_('You need to give id and text!'))
				else:
					return self.getUrl('edit_html', errmsg=_('You need to give id and text!'))
			id = _normPath(id)
			if oldid and text and oldid != id: 
				self.renameNode(oldid, id, renderNew=0)
			id = self.addNode(id, title, text, template=int(template))
			if errmsg: 
				errmsg = ' '.join([
					_("Adjusted invalid ID before saving your record: "), 
					errmsg])
				return self.getUrl('edit_html', id=str(id), errmsg=errmsg)
			else: 
				return self.getUrl('show_html', id=str(id))

	# this method renames an existing node
	def renameNode(self, id, newid, renderNew=1): 
		self.logVerbose("Renaming Wiki node %s to %s", id, newid)
		try:
			self._acquire()
			(idx, found) = self.nodes.locate({'id':id})
			if found:
				self.unrenderNode(id)
				self.nodes[idx].id = newid
				self._commit()
				if renderNew: 
					self.renderNode(newid)
				meshTool = self.getToolByNamespace('mesh')
				meshTool.renItem(id, self.name, newid)
		finally: self._release()
	
	# this method deletes an existing node
	def deleteNode(self, id): 
		self.logVerbose("Deleting Wiki node %s", id)
		try:
			self._acquire()
			(idx, found) = self.nodes.locate({'id':id})
			if found:
				self.nodes.delete(idx)
				self._commit()
				self.unrenderNode(id)
			meshTool = self.getToolByNamespace('mesh')
			meshTool.delItem(id, self)
		finally: self._release()

	# This method deletes the rendered output for a given node. 
	def _deleteRenderedNode(self, id, dummy):
		self.logVerbose("Deleting rendered output for Wiki node %s", id)
		self.deleteFile('%s.html' % id)
		mirror = self.getToolByNamespace('mirror')
		mirror.deleteItem(self, id)

	# This method delivers the description of your preferences panels
	# to hook up into the preferences list.
	def getPreferencePanels(self):
		return [] # not implemented
			
	# Internal getNode; returns the row, not a dict.
	def _getNode(self, id): 
		try:
			# self._acquire()
			(res,found) = self.nodes.locate({'id':id})
			if found: 
				return self.nodes[res]
			else: 
				return None
		finally: 
			# self._release
			pass

	# get a posting by ID
	def getNode(self, id):
		node = self._getNode(id)
		if node is None: 
			return {}
		else: 
			title = node.title
			if not node.title: 
				title = node.baseid
			return {
				'id': node.id,
				'title': title,
				'text': node.text,
				'pubtime': node.pubtime,
				'categoryid': node.categoryid, 
				'baseid': node.baseid,
				'template': node.template,
				'cloud': node.cloud,
			}

	# get a list of all template nodes
	def getTemplates(self):
		liste = []
		for node in self.nodes.select({'template':1}):
			liste.append({
				'id': node.id,
				'title': node.title,
				'pubtime': node.pubtime,
				'categoryid': node.categoryid,
				'baseid': node.baseid,
				'template': node.template,
			})
		return liste

	# get a list of all nodes (without their text!)
	def getNodes(self):
		liste = []
		for node in self.nodes:
			liste.append({
				'id': node.id,
				'title': node.title,
				'pubtime': node.pubtime,
				'categoryid': node.categoryid,
				'baseid': node.baseid,
				'template': node.template,
			})
		return liste

	# This method adds a new node to the database. Template nodes are never
	# rendered, but are given to the user for selection when a new node
	# is created.
	def addNode(self, id, title, text, template=0): 
		categoryid, baseid = os.path.split(id)
		macros = self.getToolByNamespace('macros')
		hook = ''
		renderjob = ()
		try:
			self._acquire()
			pubtime = time.time()
			(idx, found) = self.nodes.locate({'id':id})
			if found:
				node = self.nodes[idx]
				if node.title != title or node.text != text or node.categoryid != categoryid or node.baseid != baseid:
					oldtemplate = node.template
					# TODO: if the ID is unchanged, category and baseid must also be unchanged, right? 
					node.title = title
					node.text = text
					node.pubtime = pubtime
					node.categoryid = categoryid
					node.baseid = baseid
					node.template = template
					self._commit()
					if template:
						if not(oldtemplate):
							self.unrenderNode(id)
					else: 
						self.renderNode(id)
						meshTool = self.getToolByNamespace('mesh')
						renderjob = (meshTool.renderInbound, (id, self))

			else: # NEW!
				self.nodes.append({
					'id': id,
					'title': title,
					'text': text,
					'pubtime': pubtime,
					'baseid': baseid,
					'categoryid': categoryid,
					'template': template
				})
				self._commit()
				if not(template):
					self.renderNode(id)
					meshTool = self.getToolByNamespace('mesh')
					renderjob = (meshTool.renderItemsWithUnresolvedLinks, ())
			return id
		finally:
			self._release()
			if renderjob: 
				method, args = renderjob
				method(*args)
			if hook and hasattr(macros, hook):
				getattr(macros, hook)(id)

	def _rerender(self):
		count = 0
		for node in self.nodes:
			self.renderNode(node.id)
			count += 1
		return count

	# this method renders a single page in the background.
	def renderNode(self, id):
		self.logVerbose("Queueing render for Wiki node %s", id)
		self.queue.append(self.renderNodeOnline, (id,), None)
		return self.queue.length()

	# this method removes the rendered output of a node in the background
	def unrenderNode(self, id):
		self.logVerbose("Queueing unrender for Wiki node %s", id)
		self.queue.append(self._deleteRenderedNode, id, None)
		return self.queue.length()

	# this method renders a single page, but does it online (not queued)
	def renderNodeOnline(self, key, data): 
		import time
		(id,) = key
		try: 
			self.logVerbose("Rendering Wiki node %s", id)
			begin = time.time()
			result = self.__renderNodeOnline(key, data)
			duration = time.time() - begin
			self.logVerbose("%s rendered in %.3fs", id, duration)
		except:
			import traceback
			(e, d, tb) = sys.exc_info()
			msg = _('Background thread exception %s: %s') % (e, d)
			try: 
				msg += _('Exception instance __dict__: %s') % (repr(d.__dict__),)
			except: 
				pass
			msg += '<br><br>'

			for row in traceback.extract_tb(tb):
				(file, line, function, source) = row
				msg += '<br>%s[%s] in %s' % (file, line, function)
			self.logError(msg.replace('%', '%%'))

	def renderBreadCrumbs(self, id, startWithTool=0, skipLevels=0, nodiv=0): 
		nids = _splitPath(id)
		sofar = ''
		crumbs = []
		if startWithTool: 
			crumbs.append('<strong><a href="%s" class="wikicrumb">%s</a></strong>' % (
				self.getUrl('index_html'), _('Wiki')))
		levels = 0
		for nid in nids: 
			if sofar: 
				sofar = sofar + '/' + nid
			else: 
				sofar = nid
			levels = levels + 1
			if levels <= skipLevels:
				continue
			found = self._getNode(sofar) is not None
			if not _flet.desktop: 
				if found: 
					cssclass = 'wikicrumb'
					target = self.getCloudUrlForNode(sofar)
					crumbs.append('<a class="%s" href="%s">%s</a>' % (
						cssclass, target, nid))
				else: 
					cssclass = 'undefinedwikicrumb'
					crumbs.append('<span class="%s">%s</span>' % (
						cssclass, nid))
			else: 
				if found: 
					cssclass = 'wikicrumb'
					target = self.getCloudUrlForNode(sofar)
				else: 
					cssclass = 'undefinedwikicrumb'
					target = self.getUrl('show_html', id=sofar)
				crumbs.append('<a class="%s" href="%s">%s</a>' % (
					cssclass, target, nid))
		crumbhtml = '<span class="wikicrumbs">%s</span>\n' % ('&nbsp;/&nbsp;'.join(crumbs))
		if not nodiv: 
			crumbhtml = '<div id="wikicrumbsdiv">%s</div>' % crumbhtml
		return crumbhtml
	
	def subNodes(self, id): 
		"What subnodes does this node have?"
		if id: 
			spec = '/'.join((str(id), ''))
		else: 
			spec = '' # all nodes
		try:
			self._acquire()
			return [row.id for row in self.nodes if row.id.startswith(spec)]
		finally: 
			self._release()
		
	def renderSubNodes(self, id, depthLimit=0):
		idsegs = _splitPath(id)
		skip = len(idsegs)
		subnodeids = self.subNodes(id)

		base = {}
		for subnodeid in subnodeids: 
			segs = _splitPath(subnodeid)[skip:]
			cursor = base
			for seg in segs: 
				cursor = cursor.setdefault(seg, {})
		PyDS.Tool.debugOutput("RSN: %s" % repr(base))
		def spit(out, baseids, cursor, depthLimit, depth=1): 
			ids = cursor.keys()
			ids.sort()
			if len(ids): 
				out.append('<ul>')
				for id in ids:
					nbi = list(baseids) + [id]
					where = '/'.join(nbi)
					out.append('<li>')
					out.append(self.renderBreadCrumbs(where, skipLevels=len(baseids), nodiv=1))
					if depth < depthLimit: 
						spit(out, nbi, cursor[id], depthLimit, depth+1)
					out.append('</li>')
				out.append('</ul>')
		output = []
		spit(output, idsegs, base, depthLimit)
		return '<div id="subnodes">%s</div>' % '\n'.join(output)

				
		subnodes = []
		for subnodeid in subnodeids: 
			subnodes.append('<li>%s</li>' % self.renderBreadCrumbs(subnodeid, skipLevels=skip))
		if len(subnodes):
			return '<div id="subnodes"><ul>%s</ul></div>' % ''.join(subnodes)
		return '<div id="subnodes"></div>'
		
	def __renderNodeOnline(self, key, data):
		PyDS.Tool.traceargs()
		(id,) = key
		mirror = self.getToolByNamespace('mirror')
		node = self.getNode(id=id)
		if node['template']: return ''
		pagename = '%s.html' % id
		if os.sep=='\\': 
			pagename = re.sub('/',r'\\',pagename)
		req = self._getDummyRequest()
		toolname = self.name # consider ''
		body = req.renderText(
			node['text'], 
			structured = 2, # TODO: check 'structured' pref
			uniqueId = id,
			homeTool = self)
		node['cloud'] = body
		crumbhtml = self.renderBreadCrumbs(id, startWithTool=0)
		subnodeshtml = self.renderSubNodes(id, depthLimit=1)
		PyDS.Tool.debugOutput("rendering the page")
		html = crumbhtml + subnodeshtml + req.renderPage('WikiRendering', node=node)
		PyDS.Tool.debugOutput("finished rendering the page")
		# render the result into a full page
		src = req.renderPage('BaseRendering', body=html, category='', title=node.get('title', '(no title)'))
		# ... and store it
		fn = self.storeHTML(src, page=pagename, toolname=toolname)
		try:
			self._acquire()
			(idx, found) = self.nodes.locate({'id':id})
			if found:
				self.nodes[idx].cloud = body
				self._commit()
		finally: self._release()
		mirror.syncItem(self, id)
		return fn

	# this method returns the title of the RSS feed
	def getRSSTitle(self, category=''):
		macros = self.getToolByNamespace('macros')
		prefs = self.getPreferencesTool()
		title = prefs.get('private', 'title')
		if title:
			title += ': '
			title += _('Wiki')
		else:
			title = _('An Unconfigured Python Desktop Server: Wiki')
		return macros.quoteXML(title)
	
	# this method returns the description of the RSS feed
	def getRSSDescription(self, category=''):
		macros = self.getToolByNamespace('macros')
		prefs = self.getPreferencesTool()
		return macros.quoteXML(prefs.get('private', 'description'))

	def _translateItemToRSS(self, node, fullrss=0):
		macros = self.getToolByNamespace('macros')
		desc = node.text
		if node.cloud:
			desc = node.cloud
		if not(fullrss) and len(desc) > 250:
			if node.cloud:
				desc = macros.safeHTML(node.cloud[:250]) + ' ...'
			else: desc = desc[:250] + ' ...'
		item = {
			'title': macros.quoteXML(node.title),
			'link': self.getCloudUrlForNode(node.id),
			'guid': self.getCloudUrlForNode(node.id),
			'pubdate': macros.TodayFull(now=node.pubtime),
			'pubtime': node.pubtime,
			'source': '',
			'sourceurl': '',
			'categories': [],
			'comments': '',
			'trackback': '',
			'description': macros.quoteXML(desc)
		}
		return item

	# this method returns a single item as RSS item
	def getRSSItem(self, id, category='', fullrss=0):
		(idx, found) = self.nodes.locate({'id':id})
		if found and not(self.nodes[idx].template):
			item = self._translateItemToRSS(self.nodes[idx], fullrss)
			return item
		else: return {}
	
	# this method returns the stories as an item list, orderd by time
	def getRSSItems(self, category=''):
		try:
			self._acquire()
			items = []
			for node in self.nodes:
				if not(node.template):
					item = self._translateItemToRSS(node)
					items.append(item)
			items.sort(lambda a,b: -1*cmp(a['pubtime'],b['pubtime']))
			if len(items) > 20:
				items = items[:21]
			for item in items:
				del item['pubtime']
			return items
		finally: self._release()

	# this method returns hours without stories
	def getRSSSkipHours(self, category=''):
		try:
			self._acquire()
			hours = {}
			for node in self.nodes:
				if not(node.template):
					hours[time.gmtime(node.pubtime)[3]] = 1
		finally: self._release()
		nonhours = []
		for hour in range(24):
			if not(hours.has_key(hour)):
				nonhours.append(hour)
		return nonhours


__desc__ = _("This tool allows to edit wiki postings.")

PyDS.Tool.registerTool(WikiTool, 'wiki', _('Wiki'), __desc__, 150)

